# pylint: disable=C,E1101
import numpy as np
from scipy.ndimage import affine_transform


def rotate_scalar(x, rot):
    # this function works in the convention field[x, y, z]
    invrot = np.linalg.inv(rot)
    center = (np.array(x.shape) - 1) / 2
    return affine_transform(x, matrix=invrot, offset=center - np.dot(invrot, center))


def rotate_field(x, rot, R):
    # this function works in the convention field[x, y, z]
    invrot = np.linalg.inv(rot)
    y = np.empty_like(x)
    for k in range(y.shape[0]):
        center = (np.array(x.shape[1:]) - 1) / 2
        y[k] = affine_transform(x[k], matrix=invrot, offset=center - np.dot(invrot, center))

    return np.einsum("ij,j...->i...", R, y)



# pylint: disable=invalid-name, arguments-differ, missing-docstring, line-too-long, no-member, redefined-builtin, abstract-method
from functools import partial
from typing import List, Tuple
import math

import torch
from e3nn import o3, rs
from e3nn.linear_mod import KernelLinear
from e3nn.util.eval_code import eval_code
from e3nn.util.sparse import get_sparse_buffer, register_sparse_buffer


class LearnableTensorSquare(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, linear=True, allow_change_output=False, allow_zero_outputs=False):
        super().__init__()

        self.Rs_in = rs.simplify(Rs_in)
        self.Rs_out = rs.simplify(Rs_out)

        ls = [l for _, l, _ in self.Rs_out]
        selection_rule = partial(o3.selection_rule, lfilter=lambda l: l in ls)

        if linear:
            Rs_in = [(1, 0, 1)] + self.Rs_in
        else:
            Rs_in = self.Rs_in
        self.linear = linear

        Rs_ts, T = rs.tensor_square(Rs_in, selection_rule)
        register_sparse_buffer(self, 'T', T)  # [out, in1 * in2]

        ls = [l for _, l, _ in Rs_ts]
        if allow_change_output:
            self.Rs_out = [(mul, l, p) for mul, l, p in self.Rs_out if l in ls]
        elif not allow_zero_outputs:
            assert all(l in ls for _, l, _ in self.Rs_out)

        self.kernel = KernelLinear(Rs_ts, self.Rs_out)  # [out, in, w]

    def __repr__(self):
        return "{name} ({Rs_in} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=rs.format_Rs(self.Rs_in),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def forward(self, features):
        '''
        :param features: [..., channels]
        '''
        *size, n = features.size()
        features = features.reshape(-1, n)
        assert n == rs.dim(self.Rs_in)

        if self.linear:
            features = torch.cat([features.new_ones(features.shape[0], 1), features], dim=1)
            n += 1

        T = get_sparse_buffer(self, 'T')  # [out, in1 * in2]
        kernel = (T.t() @ self.kernel().T).T.reshape(rs.dim(self.Rs_out), n, n)  # [out, in1, in2]
        features = torch.einsum('zi,zj->zij', features, features)
        features = torch.einsum('kij,zij->zk', kernel, features)
        return features.reshape(*size, -1)


class LearnableTensorProduct(torch.nn.Module):
    def __init__(self, Rs_in1, Rs_in2, Rs_out, allow_change_output=False):
        super().__init__()

        self.Rs_in1 = rs.simplify(Rs_in1)
        self.Rs_in2 = rs.simplify(Rs_in2)
        self.Rs_out = rs.simplify(Rs_out)

        ls = [l for _, l, _ in self.Rs_out]
        selection_rule = partial(o3.selection_rule, lfilter=lambda l: l in ls)

        Rs_ts, T = rs.tensor_product(self.Rs_in1, self.Rs_in2, selection_rule)
        register_sparse_buffer(self, 'T', T)  # [out, in1 * in2]

        ls = [l for _, l, _ in Rs_ts]
        if allow_change_output:
            self.Rs_out = [(mul, l, p) for mul, l, p in self.Rs_out if l in ls]
        else:
            assert all(l in ls for _, l, _ in self.Rs_out)

        self.kernel = KernelLinear(Rs_ts, self.Rs_out)  # [out, in, w]

    def forward(self, features_1, features_2):
        """
        :return:         tensor [..., channel]
        """
        *size, n = features_1.size()
        features_1 = features_1.reshape(-1, n)
        assert n == rs.dim(self.Rs_in1)
        *size2, n = features_2.size()
        features_2 = features_2.reshape(-1, n)
        assert size == size2

        T = get_sparse_buffer(self, 'T')  # [out, in1 * in2]
        kernel = (T.t() @ self.kernel().T).T.reshape(rs.dim(self.Rs_out), rs.dim(self.Rs_in1), rs.dim(self.Rs_in2))  # [out, in1, in2]
        features = torch.einsum('kij,zi,zj->zk', kernel, features_1, features_2)
        return features.reshape(*size, -1)


_tensor_product_code = """
import torch

@torch.jit.script
def main(ARGSx1: torch.Tensor, x2: torch.Tensor, w: torch.Tensor) -> torch.Tensor:
    batch = x1.shape[0]
    out = x1.new_zeros((batch, DIM))
    ein = torch.einsum

CODE
    return out
"""


def WeightedTensorProduct(Rs_in1, Rs_in2, Rs_out, normalization='component', own_weight=True):
    Rs_in1 = rs.simplify(Rs_in1)
    Rs_in2 = rs.simplify(Rs_in2)
    Rs_out = rs.simplify(Rs_out)

    instr = [
        (i_1, i_2, i_out, 'uvw')
        for i_1, (_, l_1, p_1) in enumerate(Rs_in1)
        for i_2, (_, l_2, p_2) in enumerate(Rs_in2)
        for i_out, (_, l_out, p_out) in enumerate(Rs_out)
        if abs(l_1 - l_2) <= l_out <= l_1 + l_2 and p_1 * p_2 == p_out
    ]
    return CustomWeightedTensorProduct(Rs_in1, Rs_in2, Rs_out, instr, normalization, own_weight)


def GroupedWeightedTensorProduct(Rs_in1, Rs_in2, Rs_out, groups=math.inf, normalization='component', own_weight=True):
    Rs_in1 = rs.convention(Rs_in1)
    Rs_in2 = rs.convention(Rs_in2)
    Rs_out = rs.convention(Rs_out)

    groups = min(groups, min(mul for mul, _, _ in Rs_in1), min(mul for mul, _, _ in Rs_out))

    Rs_in1 = [(mul // groups + (g < mul % groups), l, p) for mul, l, p in Rs_in1 for g in range(groups)]
    Rs_out = [(mul // groups + (g < mul % groups), l, p) for mul, l, p in Rs_out for g in range(groups)]

    instr = [
        (i_1, i_2, i_out, 'uvw')
        for i_1, (_, l_1, p_1) in enumerate(Rs_in1)
        for i_2, (_, l_2, p_2) in enumerate(Rs_in2)
        for i_out, (_, l_out, p_out) in enumerate(Rs_out)
        if abs(l_1 - l_2) <= l_out <= l_1 + l_2 and p_1 * p_2 == p_out
        if i_1 % groups == i_out % groups
    ]
    return CustomWeightedTensorProduct(Rs_in1, Rs_in2, Rs_out, instr, normalization, own_weight)


class CustomWeightedTensorProduct(torch.nn.Module):
    def __init__(
            self,
            Rs_in1: rs.TY_RS_LOOSE,
            Rs_in2: rs.TY_RS_LOOSE,
            Rs_out: rs.TY_RS_LOOSE,
            instr: List[Tuple[int, int, int, str]],
            normalization: str = 'component',
            own_weight: bool = True,
            _specialized_code=True,
        ):
        """
        Create a Tensor Product operation that has each of his path weighted by a parameter.
        `instr` is a list of instructions.
        An instruction if of the form (i_1, i_2, i_out, mode)
        it means "Put `Rs_in1[i_1] otimes Rs_in2[i_2] into Rs_out[i_out]"
        `mode` determines the way the multiplicities are treated.
        The default mode should be 'uvw', meaning that all paths are created.
        """

        super().__init__()

        assert normalization in ['component', 'norm']

        self.Rs_in1 = rs.convention(Rs_in1)
        self.Rs_in2 = rs.convention(Rs_in2)
        self.Rs_out = rs.convention(Rs_out)

        code = ""

        index_w = 0
        wigners = set()
        count = [0 for _ in range(rs.dim(self.Rs_out))]

        instr = sorted(instr)  # for optimization

        for i_1, (mul_1, l_1, p_1) in enumerate(self.Rs_in1):
            index_1 = rs.dim(self.Rs_in1[:i_1])
            dim_1 = mul_1 * (2 * l_1 + 1)
            code += f"    x1_{i_1} = x1[:, {index_1}:{index_1+dim_1}].reshape(batch, {mul_1}, {2 * l_1 + 1})\n"
        code += f"\n"

        for i_2, (mul_2, l_2, p_2) in enumerate(self.Rs_in2):
            index_2 = rs.dim(self.Rs_in2[:i_2])
            dim_2 = mul_2 * (2 * l_2 + 1)
            code += f"    x2_{i_2} = x2[:, {index_2}:{index_2+dim_2}].reshape(batch, {mul_2}, {2 * l_2 + 1})\n"
        code += f"\n"

        last_ss = None

        for i_1, i_2, i_out, mode in instr:
            mul_1, l_1, p_1 = self.Rs_in1[i_1]
            mul_2, l_2, p_2 = self.Rs_in2[i_2]
            mul_out, l_out, p_out = self.Rs_out[i_out]
            dim_1 = mul_1 * (2 * l_1 + 1)
            dim_2 = mul_2 * (2 * l_2 + 1)
            dim_out = mul_out * (2 * l_out + 1)
            index_1 = rs.dim(self.Rs_in1[:i_1])
            index_2 = rs.dim(self.Rs_in2[:i_2])
            index_out = rs.dim(self.Rs_out[:i_out])

            assert p_1 * p_2 == p_out
            assert abs(l_1 - l_2) <= l_out <= l_1 + l_2

            if dim_1 == 0 or dim_2 == 0 or dim_out == 0:
                continue

            code += f"    # {l_1} x {l_2} = {l_out}\n"
            code += f"    s1 = x1_{i_1}\n"
            code += f"    s2 = x2_{i_2}\n"

            assert mode in ['uvw', 'uvu', 'uvv', 'uuw', 'uuu', 'uvuv']

            if _specialized_code:
                # optimized code for special cases:
                # 0 x 0 = 0
                # 0 x L = L
                # L x 0 = L
                # L x L = 0
                # 1 x 1 = 1

                if (l_1, l_2, l_out) == (0, 0, 0) and mode == 'uvw' and normalization in ['component', 'norm']:
                    code += f"    s1 = s1.reshape(batch, {mul_1})\n"
                    code += f"    s2 = s2.reshape(batch, {mul_2})\n"
                    dim_w = mul_1 * mul_2 * mul_out
                    code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2}, {mul_out})\n"
                    index_w += dim_w
                    code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuvw,zu,zv->zw', sw, s1, s2)\n"
                    code += "\n"

                    for pos in range(index_out, index_out + dim_out):
                        count[pos] += mul_1 * mul_2
                    continue

                if l_1 == 0 and l_2 == l_out and mode == 'uvw' and normalization == 'component':
                    code += f"    s1 = s1.reshape(batch, {mul_1})\n"
                    dim_w = mul_1 * mul_2 * mul_out
                    code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2}, {mul_out})\n"
                    index_w += dim_w
                    code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuvw,zu,zvi->zwi', sw, s1, s2).reshape(batch, {dim_out})\n"
                    code += "\n"

                    for pos in range(index_out, index_out + dim_out):
                        count[pos] += mul_1 * mul_2
                    continue

                if l_2 == 0 and l_1 == l_out and mode == 'uvw' and normalization == 'component':
                    code += f"    s2 = s2.reshape(batch, {mul_2})\n"
                    dim_w = mul_1 * mul_2 * mul_out
                    code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2}, {mul_out})\n"
                    index_w += dim_w
                    code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuvw,zui,zv->zwi', sw, s1, s2).reshape(batch, {dim_out})\n"
                    code += "\n"

                    for pos in range(index_out, index_out + dim_out):
                        count[pos] += mul_1 * mul_2
                    continue

                if l_1 == l_2 and l_out == 0 and mode == 'uvw' and normalization == 'component':
                    # Cl_l_0 = eye(3) / sqrt(2L+1)
                    dim_w = mul_1 * mul_2 * mul_out
                    code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2}, {mul_out}).div({(2 * l_1 + 1)**0.5})\n"
                    index_w += dim_w
                    code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuvw,zui,zvi->zw', sw, s1, s2).reshape(batch, {dim_out})\n"
                    code += "\n"

                    for pos in range(index_out, index_out + dim_out):
                        count[pos] += mul_1 * mul_2
                    continue

                if (l_1, l_2, l_out) == (1, 1, 1) and mode == 'uvw' and normalization == 'component':
                    # C1_1_1 = levi-civita / sqrt(2)
                    code += f"    s1 = s1.reshape(batch, {mul_1}, 1, {2 * l_1 + 1})\n"
                    code += f"    s2 = s2.reshape(batch, 1, {mul_2}, {2 * l_2 + 1})\n"
                    code += f"    s1, s2 = torch.broadcast_tensors(s1, s2)\n"
                    dim_w = mul_1 * mul_2 * mul_out
                    code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2}, {mul_out}).div({2**0.5})\n"
                    index_w += dim_w
                    code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuvw,zuvi->zwi', sw, torch.cross(s1, s2, dim=3)).reshape(batch, {dim_out})\n"
                    code += "\n"

                    for pos in range(index_out, index_out + dim_out):
                        count[pos] += mul_1 * mul_2
                    continue

            if last_ss != (i_1, i_2, mode[:2]):
                if mode[:2] == 'uv':
                    code += f"    ss = ein('zui,zvj->zuvij', s1, s2)\n"
                if mode[:2] == 'uu':
                    code += f"    ss = ein('zui,zuj->zuij', s1, s2)\n"
                last_ss = (i_1, i_2, mode[:2])

            wigners.add((l_1, l_2, l_out))

            if mode == 'uvw':
                dim_w = mul_1 * mul_2 * mul_out
                code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2}, {mul_out})\n"
                code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuvw,ijk,zuvij->zwk', sw, C{l_1}_{l_2}_{l_out}, ss).reshape(batch, {dim_out})\n"

                for pos in range(index_out, index_out + dim_out):
                    count[pos] += mul_1 * mul_2

            if mode == 'uvu':
                assert mul_1 == mul_out
                dim_w = mul_1 * mul_2
                code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2})\n"
                code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuv,ijk,zuvij->zuk', sw, C{l_1}_{l_2}_{l_out}, ss).reshape(batch, {dim_out})\n"

                for pos in range(index_out, index_out + dim_out):
                    count[pos] += mul_2

            if mode == 'uvv':
                assert mul_2 == mul_out
                dim_w = mul_1 * mul_2
                code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2})\n"
                code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuv,ijk,zuvij->zvk', sw, C{l_1}_{l_2}_{l_out}, ss).reshape(batch, {dim_out})\n"

                for pos in range(index_out, index_out + dim_out):
                    count[pos] += mul_1

            if mode == 'uuw':
                assert mul_1 == mul_2
                dim_w = mul_1 * mul_out
                code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_out})\n"
                code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuw,ijk,zuij->zwk', sw, C{l_1}_{l_2}_{l_out}, ss).reshape(batch, {dim_out})\n"

                for pos in range(index_out, index_out + dim_out):
                    count[pos] += mul_1

            if mode == 'uuu':
                assert mul_1 == mul_2 == mul_out
                dim_w = mul_1
                code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1})\n"
                code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zu,ijk,zuij->zuk', sw, C{l_1}_{l_2}_{l_out}, ss).reshape(batch, {dim_out})\n"

                for pos in range(index_out, index_out + dim_out):
                    count[pos] += 1

            if mode == 'uvuv':
                assert mul_1 * mul_2 == mul_out
                dim_w = mul_1 * mul_2
                code += f"    sw = w[:, {index_w}:{index_w+dim_w}].reshape(batch, {mul_1}, {mul_2})\n"
                code += f"    out[:, {index_out}:{index_out+dim_out}] += ein('zuv,ijk,zuvij->zuvk', sw, C{l_1}_{l_2}_{l_out}, ss).reshape(batch, {dim_out})\n"

                for pos in range(index_out, index_out + dim_out):
                    count[pos] += 1

            index_w += dim_w
            code += "\n"

        ilast = 0
        clast = count[0]
        for i, c in enumerate(count):
            if clast != c:
                if clast > 1:
                    code += f"    out[:, {ilast}:{i}].div_({clast ** 0.5})\n"
                clast = c
                ilast = i
        if clast > 1:
            code += f"    out[:, {ilast}:].div_({clast ** 0.5})\n"

        wigners = sorted(wigners)
        self.wigners_names = [f"C{l_1}_{l_2}_{l_3}" for l_1, l_2, l_3 in wigners]
        args = ", ".join(f"{arg}: torch.Tensor" for arg in self.wigners_names)
        if args:
            args += ', '

        for arg, (l_1, l_2, l_out) in zip(self.wigners_names, wigners):
            wig = o3.wigner_3j(l_1, l_2, l_out)

            if normalization == 'component':
                wig *= (2 * l_out + 1) ** 0.5
            if normalization == 'norm':
                wig *= (2 * l_1 + 1) ** 0.5 * (2 * l_2 + 1) ** 0.5

            self.register_buffer(arg, wig)

        x = _tensor_product_code
        x = x.replace("DIM", f"{rs.dim(self.Rs_out)}")
        x = x.replace("ARGS", args)
        x = x.replace("CODE", code)

        self.code = x
        self.main = eval_code(x).main
        self.nweight = index_w
        if own_weight:
            self.weight = torch.nn.Parameter(torch.randn(self.nweight))

    def __repr__(self):
        return "{name} ({Rs_in1} x {Rs_in2} -> {Rs_out} using {nw} paths)".format(
            name=self.__class__.__name__,
            Rs_in1=rs.format_Rs(self.Rs_in1),
            Rs_in2=rs.format_Rs(self.Rs_in2),
            Rs_out=rs.format_Rs(self.Rs_out),
            nw=self.nweight,
        )

    def forward(self, features_1, features_2, weight=None):
        """
        :return:         tensor [..., channel]
        """
        *size, n = features_1.size()
        features_1 = features_1.reshape(-1, n)
        assert n == rs.dim(self.Rs_in1), f"{n} is not {rs.dim(self.Rs_in1)}"
        *size2, n = features_2.size()
        features_2 = features_2.reshape(-1, n)
        assert n == rs.dim(self.Rs_in2), f"{n} is not {rs.dim(self.Rs_in2)}"
        assert size == size2

        if weight is None:
            weight = self.weight
        weight = weight.reshape(-1, self.nweight)
        if weight.shape[0] == 1:
            weight = weight.repeat(features_1.shape[0], 1)

        wigners = [getattr(self, arg) for arg in self.wigners_names]

        if features_1.shape[0] == 0:
            return torch.zeros(*size, rs.dim(self.Rs_out))

        features = self.main(*wigners, features_1, features_2, weight)
        return features.reshape(*size, -1)



# pylint: disable=arguments-differ, redefined-builtin, missing-docstring, no-member, invalid-name, line-too-long, not-callable
import collections

import torch
import torch_geometric as tg
import numpy as np
import ase.neighborlist

from e3nn import o3, rs
from e3nn.tensor import SphericalTensor


class DataNeighbors(tg.data.Data):
    """Builds a graph from points with features for message passing convolutions.

    Wraps ``_neighbor_list_and_relative_vec``; see it for edge conventions.

    Args:
        x (torch.Tensor shape [N, M]): per-node M-dimensional features.
        pos (torch.Tensor shape [N, 3]): node positions.
        r_max (float): neighbor cutoff radius.
        cell (ase.Cell/ndarray [3,3], optional): cell (box) for the points. Defaults to ``None``.
        pbc (bool or 3-tuple of bool, optional): whether to apply periodic boundary conditions to all or each of the three cell vector directions. Defaults to ``False``.
        self_interaction (bool, optional): whether to include self edges for points. Defaults to ``True``.
        **kwargs (optional): other attributes to pass to the ``torch_geometric.data.Data`` constructor.
    """
    def __init__(self, x, pos, r_max, cell=None, pbc=False, self_interaction=True, **kwargs):
        edge_index, edge_attr = _neighbor_list_and_relative_vec(
            pos,
            r_max,
            self_interaction=self_interaction,
            cell=cell,
            pbc=pbc
        )
        if cell is not None:
            # For compatability: the old DataPeriodicNeighbors put the cell
            # in the Data object as `lattice`.
            kwargs['lattice'] = cell
        super().__init__(x=x, edge_index=edge_index, edge_attr=edge_attr, pos=pos, **kwargs)

    @classmethod
    def from_ase(cls, atoms, r_max, features=None, **kwargs):
        """Build a ``DataNeighbors`` from an ``ase.Atoms`` object.

        Respects ``atoms``'s ``pbc`` and ``cell``.

        Args:
            atoms (ase.Atoms): the input.
            r_max (float): neighbor cutoff radius.
            features (torch.Tensor shape [N, M], optional): per-atom M-dimensional feature vectors. If ``None`` (the default), uses a one-hot encoding of the species present in ``atoms``.
            **kwargs (optional): other arguments for the ``DataNeighbors`` constructor.
        Returns:
            A ``DataNeighbors``.
        """
        if features is None:
            _, species_ids = np.unique(atoms.get_atomic_numbers(), return_inverse=True)
            features = torch.nn.functional.one_hot(torch.as_tensor(species_ids)).to(dtype=torch.get_default_dtype())
        return cls(
            x=features,
            pos=torch.as_tensor(atoms.positions),
            r_max=r_max,
            cell=atoms.get_cell(),
            pbc=atoms.pbc,
            **kwargs
        )


class DataPeriodicNeighbors(DataNeighbors):
    """Compatability wrapper for ``DataNeighbors``.

    Arguments are the same as ``DataNeighbors``, but ``lattice`` is accepted
    as an alias for ``cell`` and ``pbc`` is always set to ``True``.
    """
    def __init__(self, x, pos, lattice, r_max, self_interaction=True, **kwargs):
        super().__init__(
            x=x, pos=pos, cell=lattice, pbc=True, r_max=r_max,
            self_interaction=self_interaction, **kwargs
        )


class DataEdgeNeighbors(tg.data.Data):
    """Constructs graph to perform edge convolutions.

    Symmetric edges have not yet been implemented for this class.

    Args:
        x (torch.tensor shape [N, rs.dim(Rs_in)]): Node features.
        Rs_in (rs.TY_RS_STRICT): Representation list of input.
        pos (torch.tensor shape [N, 3]): Cartesian coordinates of nodes.
        r_max (float): Radial cutoff for edges.
        lmax (int > 0): Maximum L to use for SphericalTensor projection of radial distance vectors
        self_interaction (bool, optional): Include self interactions of nodes. Defaults to True.
        symmetric_edges (bool, optional): Constrain edge features to be symmetric in node index. Defaults to False.
        self_edge (float, optional): L=0 feature for self edges. Defaults to 1.
    """
    def __init__(self, x, Rs_in, pos, r_max, lmax,
                 self_interaction=True, symmetric_edges=False, self_edge=1., **kwargs):
        edge_index, edge_attr = _neighbor_list_and_relative_vec(pos, r_max, self_interaction)
        edge_index_dict, edge_edges, edge_edge_index = _get_edge_edges_and_index(edge_index, symmetric_edges=symmetric_edges)
        edge_edge_attr = []
        for _, edge2 in edge_edges:
            target2, source2 = edge2
            edge_edge_attr.append(
                pos[source2] - pos[target2]
            )

        edge_edge_index = torch.LongTensor(edge_edge_index).transpose(0, 1)
        edge_edge_attr = torch.stack(edge_edge_attr, dim=0)

        edge_x, Rs_in_edge = _initialize_edges(x, Rs_in, pos, edge_index_dict, lmax, self_edge=self_edge)

        super(DataEdgeNeighbors, self).__init__(
            x=x, edge_x=edge_x, edge_index=edge_index, edge_edge_index=edge_edge_index,
            edge_attr=edge_attr, edge_edge_attr=edge_edge_attr, pos=pos, Rs_in=Rs_in,
            Rs_in_edge=Rs_in_edge, edge_index_dict=edge_index_dict, **kwargs)


class DataEdgePeriodicNeighbors(tg.data.Data):
    """Constructs periodic graph to perform edge convolutions.

    symmetric_edges has not yet been implemented for this class.

    Args:
        x (torch.tensor shape [N, rs.dim(Rs_in)]): Node features.
        Rs_in (rs.TY_RS_STRICT): Representation list of input.
        pos (torch.tensor shape [N, 3]): Cartesian coordinates of nodes.
        lattice (torch.tensor shape [3, 3]): Lattice vectors of unit cell.
        r_max (float): Radial cutoff for edges.
        lmax (int > 0): Maximum L to use for SphericalTensor projection of radial distance vectors
        self_interaction (bool, optional): Include self interactions of nodes. Defaults to True.
        self_edge (float, optional): L=0 feature for self edges. Defaults to 1.
    """
    def __init__(self, x, Rs_in, pos, lattice, r_max, lmax,
                 self_interaction=True, self_edge=1., **kwargs):
        edge_index, edge_attr = _neighbor_list_and_relative_vec_lattice(pos, lattice, r_max, self_interaction)
        edge_index_dict, edge_edges, edge_edge_index = _get_edge_edges_and_index(edge_index, symmetric_edges=False)
        edge_edge_attr = []
        for _, edge2 in edge_edges:
            target2, source2 = edge2
            edge_edge_attr.append(
                pos[source2] - pos[target2]
            )

        edge_edge_index = torch.LongTensor(edge_edge_index).transpose(0, 1)
        edge_edge_attr = torch.stack(edge_edge_attr, dim=0)

        edge_x, Rs_in_edge = _initialize_edges(x, Rs_in, pos, edge_index_dict, lmax, self_edge=self_edge)

        super(DataEdgePeriodicNeighbors, self).__init__(
            x=x, edge_x=edge_x, edge_index=edge_index, edge_edge_index=edge_edge_index,
            edge_attr=edge_attr, edge_edge_attr=edge_edge_attr, pos=pos, Rs_in=Rs_in,
            Rs_in_edge=Rs_in_edge, edge_index_dict=edge_index_dict, **kwargs)


def _neighbor_list_and_relative_vec(pos, r_max, self_interaction=True, cell=None, pbc=False):
    """Create neighbor list and neighbor vectors based on radial cutoff.

    Create neighbor list (``edge_index``) and relative vectors
    (``edge_attr``) based on radial cutoff.

    Edges are given by the following convention:
    - ``edge_index[0]`` is the *source* (convolution center).
    - ``edge_index[1]`` is the *target* (neighbor).

    Thus, ``edge_index`` has the same convention as the relative vectors:
    :math:`\\vec{r}_{source, target}`

    Args:
        pos (shape [N, 3]): Positional coordinate; Tensor or numpy array. If Tensor, must be detached & on CPU.
        r_max (float): Radial cutoff distance for neighbor finding.
        cell (numpy shape [3, 3]): Cell for periodic boundary conditions. Ignored if ``pbc == False``.
        pbc (bool or 3-tuple of bool): Whether the system is periodic in each of the three cell dimensions.
        self_interaction (bool): Whether or not to include self-edges in the neighbor list.

    Returns:
        edge_index (torch.tensor shape [2, num_edges]): List of edges.
        edge_attr (torch.tensor shape [num_edges, 3]): Relative vectors corresponding to each edge.

    """
    if isinstance(pbc, bool):
        pbc = (pbc,)*3
    if cell is None:
        # ASE will "complete" this correctly.
        cell = np.zeros((3, 3))
    cell = ase.geometry.complete_cell(cell)

    first_idex, second_idex, displacements = ase.neighborlist.primitive_neighbor_list(
        'ijD',
        pbc,
        np.asarray(cell),
        np.asarray(pos),
        cutoff=r_max,
        self_interaction=self_interaction,
        use_scaled_positions=False
    )
    edge_index = torch.vstack((
        torch.LongTensor(first_idex),
        torch.LongTensor(second_idex)
    ))
    edge_attr = torch.as_tensor(displacements)
    return edge_index, edge_attr


def _neighbor_list_and_relative_vec_lattice(pos, lattice, r_max, self_interaction=True):
    """Create neighbor list and neighbor vectors based on radial cutoff and periodic lattice.

    Compatability wrapper around ``_neighbor_list_and_relative_vec``.
    Prefer to use ``_neighbor_list_and_relative_vec`` directly in new code.
    """
    return _neighbor_list_and_relative_vec(
        pos,
        r_max,
        cell=lattice,
        pbc=True,
        self_interaction=self_interaction
    )


def _initialize_edges(x, Rs_in, pos, edge_index_dict, lmax, self_edge=1., symmetric_edges=False):
    """Initialize edge features of DataEdgeNeighbors using node features and SphericalTensor.

    Args:
        x (torch.tensor shape [N, rs.dim(Rs_in)]): Node features.
        Rs_in (rs.TY_RS_STRICT): Representation list of input.
        pos (torch.tensor shape [N, 3]): Cartesian coordinates of nodes.
        edge_index (torch.LongTensor shape [2, num_edges]): Edges described by index of node target then node source.
        lmax (int > 0): Maximum L to use for SphericalTensor projection of radial distance vectors
        self_edge (float, optional): L=0 feature for self edges. Defaults to 1.
        symmetric_edges (bool, optional): Constrain edge features to be symmetric in node index. Defaults to False

    Returns:
        edge_x: Edge features.
        Rs_edge (rs.TY_RS_STRICT): Representation list of edge features.
    """
    edge_x = []
    if symmetric_edges:
        Rs, Q = rs.reduce_tensor('ij=ji', i=Rs_in)
    else:
        Rs, Q = rs.reduce_tensor('ij', i=Rs_in, j=Rs_in)
    Q = Q.reshape(-1, rs.dim(Rs_in), rs.dim(Rs_in))
    Rs_sph = [(1, l, (-1)**l) for l in range(lmax + 1)]
    tp_kernel = rs.TensorProduct(Rs, Rs_sph, o3.selection_rule)
    keys, values = list(zip(*edge_index_dict.items()))
    sorted_edges = sorted(zip(keys, values), key=lambda x: x[1])
    for (target, source), _ in sorted_edges:
        Ia = x[target]
        Ib = x[source]
        vector = (pos[source] - pos[target]).reshape(-1, 3)
        if torch.allclose(vector, torch.zeros(vector.shape)):
            signal = torch.zeros(rs.dim(Rs_sph))
            signal[0] = self_edge
        else:
            signal = SphericalTensor.from_geometry(vector, lmax=lmax).signal
            if symmetric_edges:
                signal += SphericalTensor.from_geometry(-vector, lmax=lmax).signal
                signal *= 0.5
        output = torch.einsum('kij,i,j->k', Q, Ia, Ib)
        output = tp_kernel(output, signal)
        edge_x.append(output)
    edge_x = torch.stack(edge_x, dim=0)
    return edge_x, tp_kernel.Rs_out


def _get_edge_edges_and_index(edge_index, symmetric_edges=False):
    """Given edge_index, construct edge_edges and edge_edge_index.

    Args:
        edge_index (torch.LongTensor shape [2, num_edges]): Edges described by index of node target then node source.
        symmetric_edges (bool, optional): Constrain edge features to be symmetric in node index. Defaults to False

    Returns:
        edge_index_dict: Dictionary of edge in terms of node indices and edge index.
        edge_edges: Pairs of edges over which to do edge convolutions using node indices. [num_edge_edges, 2, 2]
        edge_edge_index: Pairs of edges over which to do edge convolutions using edge indices. [num_edge_edges, 2]
    """
    edge_edges = []
    for target1, source1 in edge_index.transpose(1, 0).numpy():
        for target2, source2 in edge_index.transpose(1, 0).numpy():
            if target1 == target2:
                edge_edges.append(
                    [[target1, source1], [target2, source2]]
                )
    if symmetric_edges:
        distinct_edges = sorted(set(map(tuple,
                                        torch.sort(edge_index.transpose(1, 0),
                                                   dim=-1)[0].numpy().tolist())))
        print(distinct_edges)
        edge_index_dict = collections.OrderedDict(zip(distinct_edges, range(len(distinct_edges))))
        edge_edge_index = [
            [edge_index_dict[tuple(sorted(edge1))], edge_index_dict[tuple(sorted(edge2))]]
            for edge1, edge2 in edge_edges
        ]
    else:
        edge_index_dict = collections.OrderedDict(zip(map(tuple, edge_index.transpose(1, 0).numpy()), range(edge_index.shape[-1])))
        edge_edge_index = [
            [edge_index_dict[tuple(edge1)], edge_index_dict[tuple(edge2)]]
            for edge1, edge2 in edge_edges
        ]
    return edge_index_dict, edge_edges, edge_edge_index



# pylint: disable=arguments-differ, redefined-builtin, missing-docstring, no-member, invalid-name, line-too-long, not-callable
import torch

from e3nn import rs
from e3nn.non_linearities import GatedBlock, GatedBlockParity
from e3nn.non_linearities.rescaled_act import sigmoid, swish, tanh
from e3nn.linear import Linear


class DepthwiseConvolution(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, Rs_mid1, Rs_mid2, groups, convolution, linear=Linear, scalar_activation=swish, gate_activation=sigmoid, final_nonlinearity=True):
        """
        :param Rs_in:
        :param Rs_out:
        :param Rs_mid1:
        :param Rs_mid2:
        :param convolution: convolution operation that takes Rs_mid1, Rs_mid2
            e.g. convolution = lambda Rs_in, Rs_out: Convolution(Kernel(Rs_in, Rs_out, ConstantRadialModel))
        :param linear:
        :param scalar_activation:
        :param gated_activation:
        :param final_nonlinearity:
        """
        super().__init__()

        act_in = GatedBlock(
            groups * Rs_mid1, scalar_activation, gate_activation)
        self.lin_in = linear(Rs_in, act_in.Rs_in)
        self.act_in = act_in

        act_mid = GatedBlock(Rs_mid2, scalar_activation, gate_activation)
        self.conv = convolution(Rs_mid1, act_mid.Rs_in)
        self.act_mid = act_mid

        if final_nonlinearity:
            act_out = GatedBlock(Rs_out, scalar_activation, gate_activation)
            self.lin_out = linear(groups * Rs_mid2, act_out.Rs_in)
            self.act_out = act_out
        else:
            self.lin_out = linear(groups * Rs_mid2, Rs_out)
            self.act_out = None

        self.groups = groups

    def forward(self, features, *args, **kwargs):
        """
        :param features: tensor [..., point, channel]
        :return:         tensor [..., point, channel]
        """
        features = self.lin_in(features)
        features = self.act_in(features)

        features = self.conv(features, *args, **kwargs, groups=self.groups)
        features = self.act_mid(
            features.reshape(-1, rs.dim(self.act_mid.Rs_in))).reshape(*features.shape[:-1], -1)

        features = self.lin_out(features)
        if self.act_out is not None:
            features = self.act_out(features)

        return features


class DepthwiseConvolutionParity(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, Rs_mid1, Rs_mid2, groups, convolution, linear=Linear, scalar_activation=swish, gate_activation=sigmoid, final_nonlinearity=True):
        super().__init__()

        # Linear with GatedBlock
        scalars = [(mul, l, p) for (mul, l, p) in groups * Rs_mid1 if l == 0]
        act_scalars = [(mul, scalar_activation if p == 1 else tanh) for mul, l, p in scalars]

        nonscalars = [(mul, l, p) for (mul, l, p) in groups * Rs_mid1 if l > 0]
        gates = [(rs.mul_dim(nonscalars), 0, +1)]
        act_gates = [(-1, gate_activation)]

        act_in = GatedBlockParity(scalars, act_scalars, gates, act_gates, nonscalars)
        self.lin_in = linear(Rs_in, act_in.Rs_in)
        self.act_in = act_in

        # Kernel with GatedBlock
        scalars = [(mul, l, p) for (mul, l, p) in Rs_mid2 if l == 0]
        act_scalars = [(mul, scalar_activation if p == 1 else tanh) for mul, l, p in scalars]

        nonscalars = [(mul, l, p) for (mul, l, p) in Rs_mid2 if l > 0]
        gates = [(rs.mul_dim(nonscalars), 0, +1)]
        act_gates = [(-1, gate_activation)]

        act_mid = GatedBlockParity(scalars, act_scalars, gates, act_gates, nonscalars)
        self.conv = convolution(Rs_mid1, act_mid.Rs_in)
        self.act_mid = act_mid

        # Linear with or without GatedBlock
        if final_nonlinearity:
            scalars = [(mul, l, p) for (mul, l, p) in Rs_out if l == 0]
            act_scalars = [(mul, scalar_activation if p == 1 else tanh) for mul, l, p in scalars]

            nonscalars = [(mul, l, p) for (mul, l, p) in Rs_out if l > 0]
            gates = [(rs.mul_dim(nonscalars), 0, +1)]
            act_gates = [(-1, gate_activation)]

            act_out = GatedBlockParity(scalars, act_scalars, gates, act_gates, nonscalars)
            self.lin_out = linear(groups * Rs_mid2, act_out.Rs_in)
            self.act_out = act_out
        else:
            self.lin_out = linear(groups * Rs_mid2, Rs_out)
            self.act_out = None

        self.groups = groups

    def forward(self, features, *args, **kwargs):
        """
        :param features: tensor [..., point, channel]
        :return:         tensor [..., point, channel]
        """
        features = self.lin_in(features)
        features = self.act_in(features, groups=self.groups)

        features = self.conv(features, *args, **kwargs, groups=self.groups)
        features = self.act_mid(
            features.reshape(-1, rs.dim(self.act_mid.Rs_in))).reshape(*features.shape[:-1], -1)

        features = self.lin_out(features)
        if self.act_out is not None:
            features = self.act_out(features)

        return features



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, unexpected-keyword-arg, too-many-lines, redefined-builtin, arguments-differ, abstract-method
"""
Some functions related to SO3 and his usual representations

Using ZYZ Euler angles parametrisation
"""

import itertools
from functools import partial, reduce
from math import gcd
from typing import List, Tuple, Union

import torch
from torch_sparse import SparseTensor

from e3nn import o3, perm
from e3nn.util.default_dtype import torch_default_dtype
from e3nn.util.sparse import get_sparse_buffer, register_sparse_buffer

TY_RS_LOOSE = Union[List[Union[int, Tuple[int, int], Tuple[int, int, int]]], int]
TY_RS_STRICT = List[Tuple[int, int, int]]


def rep(Rs, alpha, beta, gamma, parity=None):
    """
    Representation of O(3). Parity applied (-1)**parity times.
    """
    abc = [alpha, beta, gamma]
    if parity is None:
        return o3.direct_sum(*[o3.irr_repr(l, *abc) for mul, l, _ in simplify(Rs) for _ in range(mul)])
    else:
        assert all(parity != 0 for _, _, parity in simplify(Rs))
        return o3.direct_sum(*[(p ** parity) * o3.irr_repr(l, *abc) for mul, l, p in simplify(Rs) for _ in range(mul)])


def randn(*size, normalization='component', dtype=None, device=None, requires_grad=False):
    """
    random tensor of representation Rs
    """
    di = 0
    Rs = None
    for di, n in enumerate(size):
        if isinstance(n, list):
            if Rs is not None:
                raise RuntimeError('Only one Rs is allowed')
            lsize = size[:di]
            Rs = convention(n)
            rsize = size[di + 1:]
    if Rs is None:
        raise RuntimeError('Rs missing')

    if normalization == 'component':
        return torch.randn(*lsize, dim(Rs), *rsize, dtype=dtype, device=device, requires_grad=requires_grad)
    if normalization == 'norm':
        x = torch.zeros(*lsize, dim(Rs), *rsize, dtype=dtype, device=device, requires_grad=requires_grad)
        with torch.no_grad():
            start = 0
            for mul, l, _p in Rs:
                r = torch.randn(*lsize, mul, 2 * l + 1, *rsize)
                r.div_(r.norm(2, dim=di + 1, keepdim=True))
                x.narrow(di, start, mul * (2 * l + 1)).copy_(r.reshape(*lsize, -1, *rsize))
                start += mul * (2 * l + 1)
        return x
    assert False, "normalization needs to be 'norm' or 'component'"


def haslinearpath(Rs_in: TY_RS_STRICT, l_out: int, p_out: int, selection_rule: o3.TY_SELECTION_RULE = o3.selection_rule):
    """
    :param Rs_in: list of triplet (multiplicity, representation order, parity)
    :return: if there is a linear operation between them
    """
    for mul_in, l_in, p_in in Rs_in:
        if mul_in == 0:
            continue

        for l in selection_rule(l_in, p_in, l_out, p_out):
            if p_out in (0, p_in * (-1) ** l):
                return True
    return False


def transpose_mul(Rs, cmul=-1):
    """
    :param Rs: [(mul, 0), (mul, 1), (mul, 2)]
    :return:   mul * [(1, 0), (1, 1), (1, 2)]
    """
    Rs = simplify(Rs)
    muls = {mul for mul, _, _ in Rs}
    if cmul == -1:
        cmul = reduce(gcd, muls)
    assert all(mul % cmul == 0 for mul, _, _ in Rs)

    return cmul, [(mul // cmul, l, p) for mul, l, p in Rs]


def cut(features, *Rss, dim_=-1):
    """
    Cut `feaures` according to the list of Rs
    ```
    x = rs.randn(10, Rs1 + Rs2)
    x1, x2 = cut(x, Rs1, Rs2)
    ```
    """
    index = 0
    outputs = []
    for Rs in Rss:
        n = dim(Rs)
        yield features.narrow(dim_, index, n)
        index += n
    assert index == features.shape[dim_]
    return outputs


def split_by_mul(Rs: TY_RS_LOOSE):
    """
    Example `split_by_mul([(1, 1), (2, 2)]) = [[(1, 1)], [(1, 2)], [(1, 2)]]`
    """
    Rs = simplify(Rs)

    for mul, l, p in Rs:
        for _ in range(mul):
            yield [(1, l, p)]


class TransposeToMulL(torch.nn.Module):
    """
    [(mul, 1), (mul, 2)]  ->  mul * [(1, 1), (1, 2)]
    [batch, l * mul * m]  ->  [batch, mul, l * m]
    """
    def __init__(self, Rs):
        super().__init__()
        self.Rs_in = convention(Rs)
        self.mul, self.Rs_out = transpose_mul(self.Rs_in)
        register_sparse_buffer(self, 'mixing_matrix', rearrange(self.Rs_in, self.mul * self.Rs_out))

    def __repr__(self):
        return "{name} ({Rs_in} -> {mul} x {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=format_Rs(self.Rs_in),
            mul=self.mul,
            Rs_out=format_Rs(self.Rs_out),
        )

    def forward(self, features):
        """
        ...
        """
        *size, n = features.size()
        features = features.reshape(-1, n)

        mixing_matrix = get_sparse_buffer(self, 'mixing_matrix')
        # features = torch.einsum('ij,zj->zi', self.mixing_matrix, features)
        features = (mixing_matrix @ features.T).T
        return features.reshape(*size, self.mul, -1)


def rearrange(Rs_in, Rs_out):
    """
    :return: permutation_matrix

    >>> rearrange([(1, 0), (1, 1)], [(1, 1), (1, 0)])
    tensor([[0., 1., 0., 0.],
            [0., 0., 1., 0.],
            [0., 0., 0., 1.],
            [1., 0., 0., 0.]])

    Example usage:
    permuted_input = einsum('ij,j->i', rearrange(Rs_in, Rs_out), input)
    """
    Rs_in, a = sort(Rs_in)
    Rs_out, b = sort(Rs_out)
    assert simplify(Rs_in) == simplify(Rs_out)
    return b.t() @ a


def sort(Rs: TY_RS_LOOSE) -> SparseTensor:
    """
    :return: (Rs_out, permutation_matrix)
    stable sorting of the representation by (l, p)

    sorted = perm @ unsorted

    >>> sort([(1, 1), (1, 0)])
    ([(1, 0, 0), (1, 1, 0)],
    tensor([[0., 0., 0., 1.],
            [1., 0., 0., 0.],
            [0., 1., 0., 0.],
            [0., 0., 1., 0.]]))

    Example usage:
    sortedRs, permutation_matrix = sort(Rs)
    permuted_input = einsum('ij,j->i', permutation_matrix, input)
    """
    Rs_in = simplify(Rs)
    xs = []

    j = 0  # input offset
    for mul, l, p in Rs_in:
        d = mul * (2 * l + 1)
        xs.append((l, p, mul, j, d))
        j += d

    index = []

    Rs_out = []
    i = 0  # output offset
    for l, p, mul, j, d in sorted(xs):
        Rs_out.append((mul, l, p))
        for _ in range(d):
            index.append([i, j])
            i += 1
            j += 1

    index = torch.tensor(index).T
    permutation_matrix = SparseTensor(row=index[0], col=index[1], value=torch.ones(index.shape[1]))

    return Rs_out, permutation_matrix


def irrep_dim(Rs: TY_RS_LOOSE) -> int:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: number of irreps of the representation without multiplicities
    """
    Rs = convention(Rs)
    return sum(2 * l + 1 for _, l, _ in Rs)


def mul_dim(Rs: TY_RS_LOOSE) -> int:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: number of multiplicities of the representation
    """
    Rs = convention(Rs)
    return sum(mul for mul, _, _ in Rs)


def dim(Rs: TY_RS_LOOSE) -> int:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: dimention of the representation
    """
    Rs = convention(Rs)
    return sum(mul * (2 * l + 1) for mul, l, _ in Rs)


def lmax(Rs: TY_RS_LOOSE) -> int:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: maximum l present in the signal
    """
    return max(l for mul, l, _ in convention(Rs) if mul > 0)


def convention(Rs: TY_RS_LOOSE) -> TY_RS_STRICT:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: conventional version of the same list which always includes parity
    """
    if isinstance(Rs, int):
        return [(1, Rs, 0)]

    out = []
    for r in Rs:
        if isinstance(r, int):
            mul, l, p = 1, r, 0
        elif len(r) == 2:
            (mul, l), p = r, 0
        elif len(r) == 3:
            mul, l, p = r

        assert isinstance(mul, int) and mul >= 0
        assert isinstance(l, int) and l >= 0
        assert p in [0, 1, -1]

        out.append((mul, l, p))
    return out


def simplify(Rs: TY_RS_LOOSE) -> TY_RS_STRICT:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: An equivalent list with parity = {-1, 0, 1} and neighboring orders consolidated into higher multiplicity.

    Note that simplify does not sort the Rs.
    >>> simplify([(1, 1), (1, 1), (1, 0)])
    [(2, 1, 0), (1, 0, 0)]

    Same order Rs which are seperated from each other are not combined
    >>> simplify([(1, 1), (1, 1), (1, 0), (1, 1)])
    [(2, 1, 0), (1, 0, 0), (1, 1, 0)]

    Parity is normalized to {-1, 0, 1}
    >>> simplify([(1, 1, -1), (1, 1, 50), (1, 0, 0)])
    [(1, 1, -1), (1, 1, 1), (1, 0, 0)]
    """
    out = []
    Rs = convention(Rs)
    for mul, l, p in Rs:
        if out and out[-1][1:] == (l, p):
            out[-1] = (out[-1][0] + mul, l, p)
        elif mul > 0:
            out.append((mul, l, p))
    return out


def are_equal(Rs1: TY_RS_LOOSE, Rs2: TY_RS_LOOSE) -> bool:
    """
    :param Rs1: first list of triplet (multiplicity, representation order, [parity])
    :param Rs2: second list of triplet (multiplicity, representation order, [parity])

    examples:
    Rs1 = [(1, 0), (1, 0), (1, 0)]
    Rs2 = [(3, 0)]
    are_equal(Rs1, Rs2)
    >> True

    Rs1 = [(1, 0), (1, 1), (1, 0)]
    Rs2 = [(2, 0), (1, 1)]
    are_equal(Rs1, Rs2)
    >> False
    Irreps are not in the same order
    """
    return simplify(Rs1) == simplify(Rs2)


def format_Rs(Rs: TY_RS_LOOSE) -> str:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: simplified version of the same list with the parity
    """
    Rs = convention(Rs)
    d = {
        0: "",
        1: "e",
        -1: "o",
    }
    return ",".join("{}{}{}".format("{}x".format(mul) if mul > 1 else "", l, d[p]) for mul, l, p in Rs if mul > 0)


def map_irrep_to_Rs(Rs: TY_RS_LOOSE) -> torch.Tensor:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: mapping matrix from irreps to full representation order (Rs) such
    that einsum('ij,j->i', mapping_matrix, irrep_rep) will have channel
    dimension of dim(Rs).

    examples:
    Rs = [(1, 0), (2, 1), (1, 0)] will return a matrix with 0s and 1s with
    shape [dim(Rs), 1 + 3 + 1] such there is only one 1 in each row, but
    there can be multiple 1s per column.

    Rs = [(2, 0), (2, 1)] will return a matrix with 0s and 1s with
    shape [dim(Rs), 1 + 3] such there is only one 1 in each row (rep), but
    there can be multiple 1s per column (irrep).
    """
    Rs = convention(Rs)
    mapping_matrix = torch.zeros(dim(Rs), irrep_dim(Rs))
    start_irrep = 0
    start_rep = 0
    for mult, L, _ in Rs:
        for _ in range(mult):
            irrep_slice = slice(start_irrep, start_irrep + 2 * L + 1)
            rep_slice = slice(start_rep, start_rep + 2 * L + 1)
            mapping_matrix[rep_slice, irrep_slice] = torch.eye(2 * L + 1)
            start_rep += 2 * L + 1
        start_irrep += 2 * L + 1
    return mapping_matrix  # [dim(Rs), irrep_dim(Rs)]


def map_mul_to_Rs(Rs: TY_RS_LOOSE) -> torch.Tensor:
    """
    :param Rs: list of triplet (multiplicity, representation order, [parity])
    :return: mapping matrix for multiplicity and full representation order (Rs)
    such that einsum('ij,j->i', mapping_matrix, mul_rep) will have channel
    dimension of dim(Rs)

    examples:
    Rs = [(1, 0), (2, 1), (1, 0)] will return a matrix with 0s and 1s with
    shape [dim(Rs), 1 + 2 + 1] such there is only one 1 in each row, but
    there can be multiple 1s per column.

    Rs = [(2, 0), (2, 1)] will return a matrix with 0s and 1s with
    shape [dim(Rs), 2 + 2] such there is only one 1 in each row (rep), but
    there can be multiple 1s per column (irrep).
    """
    Rs = convention(Rs)
    mapping_matrix = torch.zeros(dim(Rs), mul_dim(Rs))
    start_mul = 0
    start_rep = 0
    for mult, L, _ in Rs:
        for _ in range(mult):
            rep_slice = slice(start_rep, start_rep + 2 * L + 1)
            mapping_matrix[rep_slice, start_mul] = 1.0
            start_mul += 1
            start_rep += 2 * L + 1
    return mapping_matrix  # [dim(Rs), mul_dim(Rs)]


def tensor_product(
        input1: Union[TY_RS_LOOSE, o3.TY_SELECTION_RULE],
        input2: Union[TY_RS_LOOSE, o3.TY_SELECTION_RULE],
        output: Union[TY_RS_LOOSE, o3.TY_SELECTION_RULE],
        normalization: str = 'component',
        sorted: bool = False
) -> Tuple[TY_RS_STRICT, SparseTensor]:
    """
    Compute the matrix Q
    from Rs_out to Rs_in1 tensor product with Rs_in2

    For normalization='component',
    The set of "lines" { Q[i] }_i is orthonormal

    :return: Rs_missing, Q

    examples:
    Rs_out, Q = tensor_product_in_in(Rs_in1, Rs_in2, selection_rule)
    # Rs_in1 x Rs_in2 -> Rs_out
    torch.einsum('kij,i,j->k', Q, A, B)

    Rs_in2, Q = tensor_product_in_in(Rs_in1, selection_rule, Rs_out)
    # Rs_in1 x Rs_in2 -> Rs_out
    torch.einsum('kij,i,j->k', Q, A, B)
    """

    if callable(output):
        return _tensor_product_in_in(input1, input2, output, normalization, sorted)

    if callable(input2):
        return _tensor_product_in_out(input1, input2, output, normalization, sorted)

    if callable(input1):
        Rs_in1, Q = _tensor_product_in_out(input2, input1, output, normalization, sorted)
        # [out, in2 * in1] -> [out, in1 * in2]
        row, col, val = Q.coo()
        n1 = dim(Rs_in1)
        n2 = dim(input2)
        col = n2 * (col % n1) + col // n1
        Q = SparseTensor(row=row,
                         col=col,
                         value=val,
                         sparse_sizes=(dim(output), n1 * n2))
        return Rs_in1, Q


class TensorProduct(torch.nn.Module):
    """
    Module for tensor_product
    """
    def __init__(self, input1, input2, output, normalization='component', sorted=True):
        super().__init__()

        Rs, mat = tensor_product(input1, input2, output, normalization, sorted)

        if callable(input1):
            self.Rs_in1 = Rs
            self.Rs_in2 = convention(input2)
            self.Rs_out = convention(output)
            self._complete = 'in1'
        if callable(input2):
            self.Rs_in1 = convention(input1)
            self.Rs_in2 = Rs
            self.Rs_out = convention(output)
            self._complete = 'in2'
        if callable(output):
            self.Rs_in1 = convention(input1)
            self.Rs_in2 = convention(input2)
            self.Rs_out = Rs
            self._complete = 'out'

        register_sparse_buffer(self, 'mixing_matrix', mat)

    def __repr__(self):
        return "{name} ({Rs_in1} x {Rs_in2} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in1=format_Rs(self.Rs_in1),
            Rs_in2=format_Rs(self.Rs_in2),
            Rs_out=format_Rs(self.Rs_out),
        )

    def forward(self, features_1, features_2=None):
        '''
        :param features_1: [..., in1] or [..., in1, in2] if features_2 is None
        :param features_2: [..., in2]
        :return: [..., out]
        '''
        d_out = dim(self.Rs_out)
        d_in1 = dim(self.Rs_in1)
        d_in2 = dim(self.Rs_in2)

        if self._complete == 'out' or features_2 is None:
            if features_2 is None:
                features = features_1
            else:
                features = features_1[..., :, None] * features_2[..., None, :]

            size = features.shape[:-2]
            features = features.reshape(-1, d_in1, d_in2)  # [in1, in2, batch]

            mixing_matrix = get_sparse_buffer(self, "mixing_matrix")  # [out, in1 * in2]

            features = torch.einsum('zij->ijz', features)  # [in1, in2, batch]
            features = features.reshape(d_in1 * d_in2, features.shape[2])
            features = mixing_matrix @ features  # [out, batch]
            return features.T.reshape(*size, d_out)
        if self._complete == 'in1':
            k = self.left(features_1)  # [..., out, in2]
            return torch.einsum('...ij,...j->...i', k, features_2)
        if self._complete == 'in2':
            k = self.right(features_2)  # [..., out, in1]
            return torch.einsum('...ij,...j->...i', k, features_1)

    def right(self, features_2):
        '''
        :param features_2: [..., in2]
        :return: [..., out, in1]
        '''
        d_out = dim(self.Rs_out)
        d_in1 = dim(self.Rs_in1)
        d_in2 = dim(self.Rs_in2)
        size_2 = features_2.shape[:-1]
        if d_in2 == 0:
            return features_2.new_zeros(*size_2, d_out, d_in1)

        features_2 = features_2.reshape(-1, d_in2)

        mixing_matrix = get_sparse_buffer(self, "mixing_matrix")  # [out, in1 * in2]
        mixing_matrix = mixing_matrix.sparse_reshape(d_out * d_in1, d_in2)
        output = mixing_matrix @ features_2.T  # [out * in1, batch]
        return output.T.reshape(*size_2, d_out, d_in1)

    def left(self, features_1):
        '''
        :param features_1: [..., in1]
        :return: [..., out, in2]
        '''
        d_out = dim(self.Rs_out)
        d_in1 = dim(self.Rs_in1)
        d_in2 = dim(self.Rs_in2)
        size_1 = features_1.shape[:-1]
        if d_in1 == 0:
            return features_1.new_zeros(*size_1, d_out, d_in2)

        features_1 = features_1.reshape(-1, d_in1)

        mixing_matrix = get_sparse_buffer(self, "mixing_matrix")  # [out, in1 * in2]
        mixing_matrix = mixing_matrix.sparse_reshape(d_out * d_in1, d_in2).t()  # [in2, out * in1]
        mixing_matrix = mixing_matrix.sparse_reshape(d_in2 * d_out, d_in1)  # [in2 * out, in1]
        output = mixing_matrix @ features_1.T  # [in2 * out, batch]
        output = output.reshape(d_in2, d_out, features_1.shape[0])
        output = torch.einsum('jiz->zij', output)
        return output.reshape(*size_1, d_out, d_in2)

    def to_dense(self):
        """
        :return: tensor of shape [dim(Rs_out), dim(Rs_in1), dim(Rs_in2)]
        """
        mixing_matrix = get_sparse_buffer(self, "mixing_matrix")  # [out, in1 * in2]
        mixing_matrix = mixing_matrix.to_dense()
        mixing_matrix = mixing_matrix.reshape(dim(self.Rs_out), dim(self.Rs_in1), dim(self.Rs_in2))
        return mixing_matrix


def _tensor_product_in_in(Rs_in1, Rs_in2, selection_rule, normalization, sorted):
    """
    Compute the matrix Q
    from Rs_out to Rs_in1 tensor product with Rs_in2
    where Rs_out is a direct sum of irreducible representations

    For normalization='component',
    The set of "lines" { Q[i] }_i is orthonormal

    :return: Rs_out, Q

    example:
    _, Q = tensor_product_in_in(Rs_in1, Rs_in2)
    torch.einsum('kij,i,j->k', Q, A, B)
    """
    assert normalization in ['norm', 'component'], "normalization needs to be 'norm' or 'component'"

    Rs_in1 = simplify(Rs_in1)
    Rs_in2 = simplify(Rs_in2)

    Rs_out = []

    for mul_1, l_1, p_1 in Rs_in1:
        for mul_2, l_2, p_2 in Rs_in2:
            for l_out in selection_rule(l_1, p_1, l_2, p_2):
                Rs_out.append((mul_1 * mul_2, l_out, p_1 * p_2))

    Rs_out = simplify(Rs_out)

    dim_in2 = dim(Rs_in2)
    row = []
    col = []
    val = []

    index_out = 0

    index_1 = 0
    for mul_1, l_1, p_1 in Rs_in1:
        dim_1 = mul_1 * (2 * l_1 + 1)

        index_2 = 0
        for mul_2, l_2, p_2 in Rs_in2:
            dim_2 = mul_2 * (2 * l_2 + 1)
            for l_out in selection_rule(l_1, p_1, l_2, p_2):
                dim_out = mul_1 * mul_2 * (2 * l_out + 1)
                C = o3.wigner_3j(l_out, l_1, l_2, cached=True)
                if normalization == 'component':
                    C *= (2 * l_out + 1) ** 0.5
                if normalization == 'norm':
                    C *= (2 * l_1 + 1) ** 0.5 * (2 * l_2 + 1) ** 0.5
                I = torch.eye(mul_1 * mul_2).reshape(mul_1 * mul_2, mul_1, mul_2)
                m = torch.einsum("wuv,kij->wkuivj", I, C).reshape(dim_out, dim_1, dim_2)
                i_out, i_1, i_2 = m.nonzero(as_tuple=False).T
                i_out += index_out
                i_1 += index_1
                i_2 += index_2
                row.append(i_out)
                col.append(i_1 * dim_in2 + i_2)
                val.append(m[m != 0])

                index_out += dim_out
            index_2 += dim_2
        index_1 += dim_1

    wigner_3j_tensor = SparseTensor(
        row=torch.cat(row) if row else torch.zeros(0, dtype=torch.long),
        col=torch.cat(col) if col else torch.zeros(0, dtype=torch.long),
        value=torch.cat(val) if val else torch.zeros(0),
        sparse_sizes=(dim(Rs_out), dim(Rs_in1) * dim(Rs_in2)))

    if sorted:
        Rs_out, perm_mat = sort(Rs_out)
        Rs_out = simplify(Rs_out)
        wigner_3j_tensor = perm_mat @ wigner_3j_tensor

    return Rs_out, wigner_3j_tensor


def _tensor_product_in_out(Rs_in1, selection_rule, Rs_out, normalization, sorted):
    """
    Compute the matrix Q
    from Rs_out to Rs_in1 tensor product with Rs_in2
    where Rs_in2 is a direct sum of irreducible representations

    For normalization='component',
    The set of "lines" { Q[i] }_i is orthonormal

    :return: Rs_in2, Q

    example:
    _, Q = tensor_product_in_out(Rs_in1, Rs_out)
    torch.einsum('kij,i,j->k', Q, A, B)
    """
    assert normalization in ['norm', 'component'], "normalization needs to be 'norm' or 'component'"

    Rs_in1 = simplify(Rs_in1)
    Rs_out = simplify(Rs_out)

    Rs_in2 = []

    for mul_out, l_out, p_out in Rs_out:
        for mul_1, l_1, p_1 in Rs_in1:
            for l_2 in selection_rule(l_1, p_1, l_out, p_out):
                Rs_in2.append((mul_1 * mul_out, l_2, p_1 * p_out))

    Rs_in2 = simplify(Rs_in2)

    dim_in2 = dim(Rs_in2)
    row = []
    col = []
    val = []

    index_2 = 0

    index_out = 0
    for mul_out, l_out, p_out in Rs_out:
        dim_out = mul_out * (2 * l_out + 1)

        n_path = 0
        for mul_1, l_1, p_1 in Rs_in1:
            for l_2 in selection_rule(l_1, p_1, l_out, p_out):
                n_path += mul_1

        index_1 = 0
        for mul_1, l_1, p_1 in Rs_in1:
            dim_1 = mul_1 * (2 * l_1 + 1)
            for l_2 in selection_rule(l_1, p_1, l_out, p_out):
                if l_2 == 0:
                    assert l_out == l_1
                    l = l_1
                    dim_2 = mul_1 * mul_out
                    i_out = []
                    i_1 = []
                    i_2 = []
                    v = 0
                    for w in range(mul_out):
                        for u in range(mul_1):
                            i_out += [(2 * l + 1) * w + m for m in range(2 * l + 1)]
                            i_1 += [(2 * l + 1) * u + m for m in range(2 * l + 1)]
                            i_2 += (2 * l + 1) * [v]
                            v += 1
                    i_out = index_out + torch.tensor(i_out)
                    i_1 = index_1 + torch.tensor(i_1)
                    i_2 = index_2 + torch.tensor(i_2)
                    m = torch.ones((2 * l + 1) * dim_2) / n_path ** 0.5
                else:
                    dim_2 = mul_1 * mul_out * (2 * l_2 + 1)
                    C = o3.wigner_3j(l_out, l_1, l_2, cached=True)
                    if normalization == 'component':
                        C *= (2 * l_out + 1) ** 0.5
                    if normalization == 'norm':
                        C *= (2 * l_1 + 1) ** 0.5 * (2 * l_2 + 1) ** 0.5
                    I = torch.eye(mul_out * mul_1).reshape(mul_out, mul_1, mul_out * mul_1) / n_path ** 0.5
                    m = torch.einsum("wuv,kij->wkuivj", I, C).reshape(dim_out, dim_1, dim_2)
                    i_out, i_1, i_2 = m.nonzero(as_tuple=True)  # slow part
                    m = m[(i_out, i_1, i_2)]
                    i_out += index_out
                    i_1 += index_1
                    i_2 += index_2

                row.append(i_out)
                col.append(i_1 * dim_in2 + i_2)
                val.append(m)

                index_2 += dim_2
            index_1 += dim_1
        index_out += dim_out

    wigner_3j_tensor = SparseTensor(
        row=torch.cat(row) if row else torch.zeros(0, dtype=torch.long),
        col=torch.cat(col) if col else torch.zeros(0, dtype=torch.long),
        value=torch.cat(val) if val else torch.zeros(0),
        sparse_sizes=(dim(Rs_out), dim(Rs_in1) * dim(Rs_in2)))

    if sorted:
        Rs_in2, perm_mat = sort(Rs_in2)
        Rs_in2 = simplify(Rs_in2)
        # sorted = perm_mat @ unsorted
        wigner_3j_tensor = wigner_3j_tensor.sparse_reshape(-1, dim(Rs_in2))
        wigner_3j_tensor = wigner_3j_tensor @ perm_mat.t()  # slow part
        wigner_3j_tensor = wigner_3j_tensor.sparse_reshape(-1, dim(Rs_in1) * dim(Rs_in2))

    return Rs_in2, wigner_3j_tensor


def tensor_square(
        Rs_in: TY_RS_LOOSE,
        selection_rule: o3.TY_SELECTION_RULE = o3.selection_rule,
        normalization: str = 'component',
        sorted: bool = False
) -> Tuple[TY_RS_STRICT, SparseTensor]:
    """
    Compute the matrix Q
    from Rs_out to Rs_in tensor product with Rs_in
    where Rs_out is a direct sum of irreducible representations

    For normalization='component',
    The set of "lines" { Q[i] }_i is orthonormal

    :return: Rs_out, Q

    example:
    _, Q = tensor_square(Rs_in)
    torch.einsum('kij,i,j->k', Q, A, A)
    """
    assert normalization in ['norm', 'component'], "normalization needs to be 'norm' or 'component'"

    Rs_in = simplify(Rs_in)

    Rs_out = []

    for i, (mul_1, l_1, p_1) in enumerate(Rs_in):
        for l_out in selection_rule(l_1, p_1, l_1, p_1):
            if l_out % 2 == 0:
                Rs_out.append((mul_1 * (mul_1 + 1) // 2, l_out, p_1**2))
            else:
                Rs_out.append((mul_1 * (mul_1 - 1) // 2, l_out, p_1**2))

        for mul_2, l_2, p_2 in Rs_in[i + 1:]:
            for l_out in selection_rule(l_1, p_1, l_2, p_2):
                Rs_out.append((mul_1 * mul_2, l_out, p_1 * p_2))

    Rs_out = simplify(Rs_out)

    dim_in = dim(Rs_in)
    row = []
    col = []
    val = []

    index_out = 0

    index_1 = 0
    for i, (mul_1, l_1, p_1) in enumerate(Rs_in):
        dim_1 = mul_1 * (2 * l_1 + 1)

        for l_out in selection_rule(l_1, p_1, l_1, p_1):
            I = torch.eye(mul_1**2).reshape(mul_1**2, mul_1, mul_1)
            uv = I.nonzero(as_tuple=False)[:, 1:]
            if l_out % 2 == 0:
                I = I[uv[:, 0] <= uv[:, 1]]
            else:
                I = I[uv[:, 0] < uv[:, 1]]

            if I.shape[0] == 0:
                continue

            C = o3.wigner_3j(l_out, l_1, l_1)
            if normalization == 'component':
                C *= (2 * l_out + 1) ** 0.5
            if normalization == 'norm':
                C *= (2 * l_1 + 1) ** 0.5 * (2 * l_1 + 1) ** 0.5
            dim_out = I.shape[0] * (2 * l_out + 1)
            m = torch.einsum("wuv,kij->wkuivj", I, C).reshape(dim_out, dim_1, dim_1)
            i_out, i_1, i_2 = m.nonzero(as_tuple=False).T
            i_out += index_out
            i_1 += index_1
            i_2 += index_1
            row.append(i_out)
            col.append(i_1 * dim_in + i_2)
            val.append(m[m != 0])

            index_out += dim_out

        index_2 = index_1 + dim_1
        for mul_2, l_2, p_2 in Rs_in[i + 1:]:
            dim_2 = mul_2 * (2 * l_2 + 1)
            for l_out in selection_rule(l_1, p_1, l_2, p_2):
                I = torch.eye(mul_1 * mul_2).reshape(mul_1 * mul_2, mul_1, mul_2)

                C = o3.wigner_3j(l_out, l_1, l_2)
                if normalization == 'component':
                    C *= (2 * l_out + 1) ** 0.5
                if normalization == 'norm':
                    C *= (2 * l_1 + 1) ** 0.5 * (2 * l_2 + 1) ** 0.5
                dim_out = I.shape[0] * (2 * l_out + 1)
                m = torch.einsum("wuv,kij->wkuivj", I, C).reshape(dim_out, dim_1, dim_2)
                i_out, i_1, i_2 = m.nonzero(as_tuple=False).T
                i_out += index_out
                i_1 += index_1
                i_2 += index_2
                row.append(i_out)
                col.append(i_1 * dim_in + i_2)
                val.append(m[m != 0])

                index_out += dim_out
            index_2 += dim_2
        index_1 += dim_1

    wigner_3j_tensor = SparseTensor(
        row=torch.cat(row) if row else torch.zeros(0, dtype=torch.long),
        col=torch.cat(col) if col else torch.zeros(0, dtype=torch.long),
        value=torch.cat(val) if val else torch.zeros(0),
        sparse_sizes=(dim(Rs_out), dim(Rs_in) * dim(Rs_in)))

    if sorted:
        Rs_out, perm_mat = sort(Rs_out)
        Rs_out = simplify(Rs_out)
        # sorted = perm_mat @ unsorted
        wigner_3j_tensor = perm_mat @ wigner_3j_tensor

    return Rs_out, wigner_3j_tensor


class TensorSquare(torch.nn.Module):
    """
    Module for tensor_square
    """
    def __init__(self, Rs_in, selection_rule=o3.selection_rule):
        super().__init__()

        self.Rs_in = simplify(Rs_in)

        self.Rs_out, mixing_matrix = tensor_square(self.Rs_in, selection_rule, sorted=True)
        register_sparse_buffer(self, 'mixing_matrix', mixing_matrix)

    def __repr__(self):
        return "{name} ({Rs_in} ^ 2 -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=format_Rs(self.Rs_in),
            Rs_out=format_Rs(self.Rs_out),
        )

    def forward(self, features):
        '''
        :param features: [..., channels]
        '''
        *size, n = features.size()
        features = features.reshape(-1, n)

        mixing_matrix = get_sparse_buffer(self, "mixing_matrix")

        features = torch.einsum('zi,zj->ijz', features, features)
        features = mixing_matrix @ features.reshape(-1, features.shape[2])
        return features.T.reshape(*size, -1)


def elementwise_tensor_product(
        Rs_in1: TY_RS_LOOSE,
        Rs_in2: TY_RS_LOOSE,
        selection_rule: o3.TY_SELECTION_RULE = o3.selection_rule,
        normalization: str = 'component'
) -> Tuple[TY_RS_STRICT, SparseTensor]:
    """
    :return: Rs_out, matrix

    m_kij A_i B_j
    """
    assert normalization in ['norm', 'component'], "normalization needs to be 'norm' or 'component'"

    Rs_in1 = simplify(Rs_in1)
    Rs_in2 = simplify(Rs_in2)

    assert sum(mul for mul, _, _ in Rs_in1) == sum(mul for mul, _, _ in Rs_in2)

    i = 0
    while i < len(Rs_in1):
        mul_1, l_1, p_1 = Rs_in1[i]
        mul_2, l_2, p_2 = Rs_in2[i]

        if mul_1 < mul_2:
            Rs_in2[i] = (mul_1, l_2, p_2)
            Rs_in2.insert(i + 1, (mul_2 - mul_1, l_2, p_2))

        if mul_2 < mul_1:
            Rs_in1[i] = (mul_2, l_1, p_1)
            Rs_in1.insert(i + 1, (mul_1 - mul_2, l_1, p_1))
        i += 1

    Rs_out = []
    for (mul, l_1, p_1), (mul_2, l_2, p_2) in zip(Rs_in1, Rs_in2):
        assert mul == mul_2
        for l in selection_rule(l_1, p_1, l_2, p_2):
            Rs_out.append((mul, l, p_1 * p_2))

    Rs_out = simplify(Rs_out)

    dim_in2 = dim(Rs_in2)
    row = []
    col = []
    val = []

    index_out = 0
    index_1 = 0
    index_2 = 0
    for (mul, l_1, p_1), (mul_2, l_2, p_2) in zip(Rs_in1, Rs_in2):
        assert mul == mul_2
        dim_1 = mul * (2 * l_1 + 1)
        dim_2 = mul * (2 * l_2 + 1)

        for l_out in selection_rule(l_1, p_1, l_2, p_2):
            dim_out = mul * (2 * l_out + 1)
            C = o3.wigner_3j(l_out, l_1, l_2, cached=True)
            if normalization == 'component':
                C *= (2 * l_out + 1) ** 0.5
            if normalization == 'norm':
                C *= (2 * l_1 + 1) ** 0.5 * (2 * l_2 + 1) ** 0.5
            I = torch.einsum("uv,wu->wuv", torch.eye(mul), torch.eye(mul))
            m = torch.einsum("wuv,kij->wkuivj", I, C).reshape(dim_out, dim_1, dim_2)
            i_out, i_1, i_2 = m.nonzero(as_tuple=False).T
            i_out += index_out
            i_1 += index_1
            i_2 += index_2
            row.append(i_out)
            col.append(i_1 * dim_in2 + i_2)
            val.append(m[m != 0])

            index_out += dim_out

        index_1 += dim_1
        index_2 += dim_2

    wigner_3j_tensor = SparseTensor(
        row=torch.cat(row) if row else torch.zeros(0, dtype=torch.long),
        col=torch.cat(col) if col else torch.zeros(0, dtype=torch.long),
        value=torch.cat(val) if val else torch.zeros(0),
        sparse_sizes=(dim(Rs_out), dim(Rs_in1) * dim(Rs_in2)))

    return Rs_out, wigner_3j_tensor


class ElementwiseTensorProduct(torch.nn.Module):
    """
    Module for elementwise_tensor_product
    """
    def __init__(self, Rs_in1, Rs_in2, selection_rule=o3.selection_rule):
        super().__init__()

        self.Rs_in1 = simplify(Rs_in1)
        self.Rs_in2 = simplify(Rs_in2)
        assert sum(mul for mul, _, _ in self.Rs_in1) == sum(mul for mul, _, _ in self.Rs_in2)

        self.Rs_out, mixing_matrix = elementwise_tensor_product(self.Rs_in1, self.Rs_in2, selection_rule)
        register_sparse_buffer(self, "mixing_matrix", mixing_matrix)

    def forward(self, features_1, features_2):
        '''
        :param features_1: [..., in1]
        :param features_2: [..., in2]
        :return: [..., out]
        '''
        d_out = dim(self.Rs_out)
        d_in1 = dim(self.Rs_in1)
        d_in2 = dim(self.Rs_in2)

        features = features_1[..., :, None] * features_2[..., None, :]

        size = features.shape[:-2]
        features = features.reshape(-1, d_in1, d_in2)  # [in1, in2, batch]

        mixing_matrix = get_sparse_buffer(self, "mixing_matrix")  # [out, in1 * in2]

        features = torch.einsum('zij->ijz', features)  # [in1, in2, batch]
        features = features.reshape(d_in1 * d_in2, features.shape[2])
        features = mixing_matrix @ features  # [out, batch]
        return features.T.reshape(*size, d_out)


def _is_representation(D, eps, with_parity=False):
    e = (0, 0, 0, 0) if with_parity else (0, 0, 0)
    I = D(*e)
    if not torch.allclose(I, I @ I):
        return False

    g1 = o3.rand_angles() + (0,) if with_parity else o3.rand_angles()
    g2 = o3.rand_angles() + (0,) if with_parity else o3.rand_angles()

    g12 = o3.compose_with_parity(*g1, *g2) if with_parity else o3.compose(*g1, *g2)
    D12 = D(*g12)

    D1D2 = D(*g1) @ D(*g2)

    return (D12 - D1D2).abs().max().item() < eps * D12.abs().max().item()


def _round_sqrt(x, eps):
    # round off x assuming it contains terms of the form +-1/sqrt(N)
    x[x.abs() < eps] = 0
    x = x.sign() / x.pow(2)
    x = x.div(eps).round().mul(eps)
    x = x.sign() / x.abs().sqrt()
    x[torch.isnan(x)] = 0
    x[torch.isinf(x)] = 0
    return x


def reduce_tensor(formula, eps=1e-9, has_parity=None, **kw_Rs):
    """
    Usage
    Rs, Q = rs.reduce_tensor('ijkl=jikl=ikjl=ijlk', i=[(1, 1)])
    Rs = 0,2,4
    Q = tensor of shape [15, 81]
    """
    dtype = torch.get_default_dtype()
    with torch_default_dtype(torch.float64):
        # reformat `formulas` and make checks
        formulas = [
            (-1 if f.startswith('-') else 1, f.replace('-', ''))
            for f in formula.split('=')
        ]
        s0, f0 = formulas[0]
        assert s0 == 1

        for _s, f in formulas:
            if len(set(f)) != len(f) or set(f) != set(f0):
                raise RuntimeError(f'{f} is not a permutation of {f0}')
            if len(f0) != len(f):
                raise RuntimeError(f'{f0} and {f} don\'t have the same number of indices')

        # `formulas` is a list of (sign, permutation of indices)
        # each formula can be viewed as a permutation of the original formula
        formulas = {(s, tuple(f.index(i) for i in f0)) for s, f in formulas}  # set of generators (permutations)

        # they can be composed, for instance if you have ijk=jik=ikj
        # you also have ijk=jki
        # applying all possible compositions creates an entire group
        while True:
            n = len(formulas)
            formulas = formulas.union([(s, perm.inverse(p)) for s, p in formulas])
            formulas = formulas.union([
                (s1 * s2, perm.compose(p1, p2))
                for s1, p1 in formulas
                for s2, p2 in formulas
            ])
            if len(formulas) == n:
                break  # we break when the set is stable => it is now a group \o/

        # lets clean the `kw_Rs` before checking that they are compatible with the formulas
        for i in kw_Rs:
            if not callable(kw_Rs[i]):
                Rs = convention(kw_Rs[i])
                if has_parity is None:
                    has_parity = any(p != 0 for _, _, p in Rs)
                if not has_parity and not all(p == 0 for _, _, p in Rs):
                    raise RuntimeError(f'{format_Rs(Rs)} parity has to be specified everywhere or nowhere')
                if has_parity and any(p == 0 for _, _, p in Rs):
                    raise RuntimeError(f'{format_Rs(Rs)} parity has to be specified everywhere or nowhere')
                kw_Rs[i] = Rs

        if has_parity is None:
            raise RuntimeError(f'please specify the argument `has_parity`')

        # here we check that each index has one and only one representation
        for _s, p in formulas:
            f = "".join(f0[i] for i in p)
            for i, j in zip(f0, f):
                if i in kw_Rs and j in kw_Rs and kw_Rs[i] != kw_Rs[j]:
                    raise RuntimeError(f'Rs of {i} (Rs={format_Rs(kw_Rs[i])}) and {j} (Rs={format_Rs(kw_Rs[j])}) should be the same')
                if i in kw_Rs:
                    kw_Rs[j] = kw_Rs[i]
                if j in kw_Rs:
                    kw_Rs[i] = kw_Rs[j]

        for i in f0:
            if i not in kw_Rs:
                raise RuntimeError(f'index {i} has not Rs associated to it')

        e = (0, 0, 0, 0) if has_parity else (0, 0, 0)
        dims = {i: len(kw_Rs[i](*e)) if callable(kw_Rs[i]) else dim(kw_Rs[i]) for i in f0}  # dimension of each index
        full_base = list(itertools.product(*(range(dims[i]) for i in f0)))  # (0, 0, 0), (0, 0, 1), (0, 0, 2), ... (3, 3, 3)
        # len(full_base) degrees of freedom in an unconstrained tensor

        # but there is constraints given by the group `formulas`
        # For instance if `ij=-ji`, then 00=-00, 01=-01 and so on
        base = set()
        for x in full_base:
            # T[x] is a coefficient of the tensor T and is related to other coefficient T[y]
            # if x and y are related by a formula
            xs = {(s, tuple(x[i] for i in p)) for s, p in formulas}
            # s * T[x] are all equal for all (s, x) in xs
            # if T[x] = -T[x] it is then equal to 0 and we lose this degree of freedom
            if not (-1, x) in xs:
                # the sign is arbitrary, put both possibilities
                base.add(frozenset({
                    frozenset(xs),
                    frozenset({(-s, x) for s, x in xs})
                }))

        # len(base) is the number of degrees of freedom in the tensor.
        # Now we want to decompose these degrees of freedom into irreps

        base = sorted([sorted([sorted(xs) for xs in x]) for x in base])  # requested for python 3.7 but not for 3.8 (probably a bug in 3.7)

        # First we compute the change of basis (projection) between full_base and base
        d_sym = len(base)
        d = len(full_base)
        Q = torch.zeros(d_sym, d)

        for i, x in enumerate(base):
            x = max(x, key=lambda xs: sum(s for s, x in xs))
            for s, e in x:
                j = full_base.index(e)
                Q[i, j] = s / len(x)**0.5

        assert torch.allclose(Q @ Q.T, torch.eye(d_sym))

        if d_sym == 0:
            return [], torch.zeros(d_sym, d).to(dtype=dtype)

        # We project the representation on the basis `base`
        def representation(alpha, beta, gamma, parity=None):
            def re(r):
                if callable(r):
                    if has_parity:
                        return r(alpha, beta, gamma, parity)
                    return r(alpha, beta, gamma)
                return rep(r, alpha, beta, gamma, parity)

            m = o3.kron(*(re(kw_Rs[i]) for i in f0))
            return Q @ m @ Q.T

        # And check that after this projection it is still a representation
        assert _is_representation(representation, eps, has_parity)

        # The rest of the code simply extract the irreps present in this representation
        Rs_out = []
        A = Q.clone()
        for l in range(int((d_sym - 1) // 2) + 1):
            for p in [-1, 1] if has_parity else [0]:
                if 2 * l + 1 > d_sym - dim(Rs_out):
                    break

                mul, B, representation = o3.reduce(representation, partial(rep, [(1, l, p)]), eps, has_parity)
                A = o3.direct_sum(torch.eye(d_sym - B.shape[0]), B) @ A
                A = _round_sqrt(A, eps)
                Rs_out += [(mul, l, p)]

                if dim(Rs_out) == d_sym:
                    break

        if dim(Rs_out) != d_sym:
            raise RuntimeError(f'unable to decompose into irreducible representations')

        return simplify(Rs_out), A.to(dtype=dtype)



"""
Evaluate a python string as code
"""
import importlib.machinery
import importlib.util
import tempfile
import os


def eval_code(code):
    """
    save code in a temporary file and import it as a module
    """
    new_file, filename = tempfile.mkstemp(text=True)

    os.write(new_file, bytes(code, 'ascii'))
    os.close(new_file)

    loader = importlib.machinery.SourceFileLoader('main', filename)
    spec = importlib.util.spec_from_loader(loader.name, loader)
    mod = importlib.util.module_from_spec(spec)
    loader.exec_module(mod)
    return mod



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, missing-docstring, arguments-differ
import math
from math import pi

import scipy.signal
import torch

from e3nn import o3, rs, rsh
from e3nn.s2grid import FromS2Grid, ToS2Grid
from e3nn.tensor.irrep_tensor import IrrepTensor


def spherical_harmonics_dirac(vectors, lmax):
    """
    approximation of a signal that is 0 everywhere except on the angle (alpha, beta) where it is one.
    the higher is lmax the better is the approximation
    """
    return 4 * pi / (lmax + 1)**2 * rsh.spherical_harmonics_xyz(list(range(lmax + 1)), vectors)


def projection(vectors, lmax):
    """
    :param vectors: tensor of shape [..., xyz]
    :return: tensor of shape [..., l * m]
    """
    coeff = spherical_harmonics_dirac(vectors, lmax)  # [..., l * m]
    radii = vectors.norm(2, dim=-1, keepdim=True)  # [...]
    return coeff * radii


def _find_peaks_2d(x):
    iii = []
    for i in range(x.shape[0]):
        jj, _ = scipy.signal.find_peaks(x[i, :])
        iii += [(i, j) for j in jj]

    jjj = []
    for j in range(x.shape[1]):
        ii, _ = scipy.signal.find_peaks(x[:, j])
        jjj += [(i, j) for i in ii]

    return list(set(iii).intersection(set(jjj)))


def adjusted_projection(vectors, lmax):
    """
    :param vectors: tensor of shape [..., xyz]
    :return: tensor of shape [l * m]
    """
    vectors = vectors.reshape(-1, 3)
    radii = vectors.norm(2, -1)  # [batch]
    vectors = vectors[radii > 0]  # [batch, 3]

    coeff = rsh.spherical_harmonics_xyz(list(range(lmax + 1)), vectors)  # [batch, l * m]
    A = torch.einsum(
        "ai,bi->ab",
        coeff,
        coeff
    )
    # Y(v_a) . Y(v_b) solution_b = radii_a
    solution = torch.lstsq(radii, A).solution.reshape(-1)  # [b]
    assert (radii - A @ solution).abs().max() < 1e-5 * radii.abs().max()

    return solution @ coeff


class SphericalTensor:
    def __init__(self, signal: torch.Tensor, p_val: int = 0, p_arg: int = 0):
        """
        f: s2 -> R

        Rotations
        [D(g) f](x) = f(g^{-1} x)

        Parity
        [P f](x) = p_val f(p_arg x)

        f(x) = sum F^l . Y^l(x)

        This class contains the F^l

        Rotations
        [D(g) f](x) = sum [D^l(g) F^l] . Y^l(x)         (using equiv. of Y and orthogonality of D)

        Parity
        [P f](x) = sum [p_val p_arg^l F^l] . Y^l(x)     (using parity of Y)
        """
        lmax = round(math.sqrt(signal.shape[-1]) - 1)

        if signal.shape[-1] != (lmax + 1)**2:
            raise ValueError(
                "Last tensor dimension and Rs do not have same dimension.")

        self.signal = signal
        self.lmax = lmax
        self.Rs = rs.convention([(1, l, p_val * p_arg**l) for l in range(lmax + 1)])
        self.p_val = p_val
        self.p_arg = p_arg

    @classmethod
    def from_geometry(cls, vectors, lmax, p=0, adjusted=True):
        """
        :param vectors: tensor of vectors (p=-1) or pseudovectors (p=1) of shape [..., 3=xyz]
        """
        if adjusted:
            signal = adjusted_projection(vectors, lmax)
        else:
            vectors = vectors.reshape(-1, 3)
            r = vectors.norm(dim=1)
            sh = rsh.spherical_harmonics_xyz(list(range(lmax + 1)), vectors)
            # 0.5 * sum_a ( Y(v_a) . sum_b r_b Y(v_b) s - r_a )^2
            A = torch.einsum('ai,b,bi->a', sh, r, sh)
            # 0.5 * sum_a ( A_a s - r_a )^2
            # sum_a A_a^2 s = sum_a A_a r_a
            s = torch.dot(A, r) / A.norm().pow(2)
            signal = s * torch.einsum('a,ai->i', r, sh)
        return cls(signal, p_val=1, p_arg=p)

    @classmethod
    def from_samples(cls, positions, values, lmax, res=100, p_val=0, p_arg=0):
        """
        :param positions: tensor of shape [num_points, 3=xyz]
        :param values: tensor of shape [num_points]
        """
        positions = positions.reshape(-1, 3)
        values = values.reshape(-1)
        positions /= positions.norm(p=2, dim=1, keepdim=True)
        assert positions.shape[0] == values.shape[0], "positions and values must have the same number of points"

        s2 = FromS2Grid(res=res, lmax=lmax, normalization='none')
        pos = s2.grid

        cd = torch.cdist(pos, positions, p=2)
        val = values[cd.argmin(2)]

        return cls(s2(val), p_val=p_val, p_arg=p_arg)

    @classmethod
    def spherical_harmonic(cls, l, m, lmax=None):
        if lmax is None:
            lmax = l
        signal = torch.zeros((lmax + 1)**2)
        signal[l**2 + l + m] = 1
        return cls(signal)

    def __repr__(self):
        p_str = ""
        if self.p_arg != 0 and self.p_val != 0:
            p_str = f", [Parity f](x) = {'-' if self.p_val == -1 else ''}f({'-' if self.p_arg == -1 else ''}x)"
        return f"{self.__class__.__name__}(lmax={self.lmax}{p_str})"

    def sph_norm(self):
        i = 0
        norms = []
        for l in range(self.lmax + 1):
            n = self.signal[..., i: i + 2 * l + 1].norm(p=2, dim=-1)
            norms.append(n)
            i += 2 * l + 1
        return torch.stack(norms, dim=-1)

    def signal_xyz(self, r):
        """
        Evaluate the signal on the sphere
        """
        sh = rsh.spherical_harmonics_xyz(list(range(self.lmax + 1)), r)
        dim = (self.lmax + 1)**2
        output = torch.einsum(
            'ai,zi->za', sh.reshape(-1, dim), self.signal.reshape(-1, dim))
        return output.reshape((*self.signal.shape[:-1], *r.shape[:-1]))

    def signal_alpha_beta(self, alpha, beta):
        """
        Evaluate the signal on the sphere
        """
        sh = rsh.spherical_harmonics_alpha_beta(
            list(range(self.lmax + 1)), alpha, beta)
        dim = (self.lmax + 1)**2
        output = torch.einsum(
            'ai,zi->za', sh.reshape(-1, dim), self.signal.reshape(-1, dim))
        return output.reshape((*self.signal.shape[:-1], *alpha.shape))

    def signal_on_grid(self, res=100, normalization='none'):
        """
        :return: [..., beta, alpha]
        Evaluate the signal on the sphere
        """
        s2 = ToS2Grid(self.lmax, res=res, normalization=normalization)
        return s2.grid, s2(self.signal)

    def plotly_surface(self, res=100, radius=True, center=None, relu=False, normalization='none'):
        """
        To use as follow
        ```
        import plotly.graph_objects as go
        surface = go.Surface(**self.plotly_surface())
        fig = go.Figure(data=[surface])
        fig.show()
        ```
        """
        r, f = self.plot(res, radius, center, relu, normalization)
        return dict(
            x=r[:, :, 0].numpy(),
            y=r[:, :, 1].numpy(),
            z=r[:, :, 2].numpy(),
            surfacecolor=f.numpy(),
        )

    def plot(self, res=100, radius=True, center=None, relu=False, normalization='none'):
        """
        Returns `(r, f)` of shapes `[beta, alpha, 3]` and `[beta, alpha]`
        """
        assert self.signal.dim() == 1

        r, f = self.signal_on_grid(res, normalization)
        f = f.relu() if relu else f

        # beta: [0, pi]
        r[0] = r.new_tensor([0, 0, 1])
        r[-1] = r.new_tensor([0, 0, -1])
        f[0] = f[0].mean()
        f[-1] = f[-1].mean()

        # alpha: [0, 2pi]
        r = torch.cat([r, r[:, :1]], dim=1)  # [beta, alpha, 3]
        f = torch.cat([f, f[:, :1]], dim=1)  # [beta, alpha]

        if radius:
            r *= f.abs().unsqueeze(-1)

        if center is not None:
            r += center

        return r, f

    def change_lmax(self, lmax):
        new_Rs = [(1, l) for l in range(lmax + 1)]
        if self.lmax == lmax:
            return self
        elif self.lmax > lmax:
            new_signal = self.signal[..., :rs.dim(new_Rs)]
            return SphericalTensor(new_signal, self.p_val, self.p_arg)
        elif self.lmax < lmax:
            new_signal = torch.zeros(*self.signal.shape[:-1], rs.dim(new_Rs))
            new_signal[..., :rs.dim(self.Rs)] = self.signal
            return SphericalTensor(new_signal, self.p_val, self.p_arg)

    def __add__(self, other):
        lmax = max(self.lmax, other.lmax)
        new_self = self.change_lmax(lmax)
        new_other = other.change_lmax(lmax)

        return SphericalTensor(new_self.signal + new_other.signal, self.p_val, self.p_arg)

    def __mul__(self, other):
        # Dot product if Rs of both objects match
        lmax = max(self.lmax, other.lmax)
        new_self = self.change_lmax(lmax)
        new_other = other.change_lmax(lmax)

        mult = new_self.signal * new_other.signal
        mapping_matrix = rs.map_mul_to_Rs(new_self.Rs)
        scalars = torch.einsum('rm,...r->...m', mapping_matrix, mult)
        Rs = [(1, 0, p1 * p2) for (_, l1, p1), (_, l2, p2) in zip(new_self.Rs, new_other.Rs)]
        return IrrepTensor(scalars, Rs)

    def dot(self, other):
        return (self * other).tensor.sum(-1)

    def __matmul__(self, other):
        # Tensor product
        # Better handle mismatch of features indices
        tp = rs.TensorProduct(self.Rs, other.Rs, o3.selection_rule)
        return IrrepTensor(tp(self.signal, other.signal), tp.Rs_out)

    @classmethod
    def from_irrep_tensor(cls, irrep_tensor):
        Rs_remove_p = [(mul, L) for mul, L, p in irrep_tensor.Rs]
        Rs, perm = rs.sort(Rs_remove_p)
        Rs = rs.simplify(Rs)
        mul, Ls, _ = zip(*Rs)
        if max(mul) > 1:
            raise ValueError(
                "Cannot have multiplicity greater than 1 for any L. This tensor has a simplified Rs of {}".format(Rs)
            )
        Lmax = max(Ls)
        sorted_tensor = torch.einsum('ij,...j->...i', perm.to_dense(), irrep_tensor.tensor)
        signal = torch.zeros((Lmax + 1)**2)
        Rs_idx = 0
        for L in range(Lmax + 1):
            if Rs[Rs_idx][1] == L:
                ten_slice = slice(rs.dim(Rs[:Rs_idx]), rs.dim(Rs[:Rs_idx + 1]))
                signal[L ** 2: (L + 1) ** 2] = sorted_tensor[ten_slice]
                Rs_idx += 1
        return cls(signal)

    def find_peaks(self, res=100):
        x1, f1 = self.signal_on_grid(res)

        abc = pi / 2, pi / 2, pi / 2
        R = o3.rot(*abc)
        D = rs.rep(self.Rs, *abc)

        rtensor = SphericalTensor(D @ self.signal)
        rx2, f2 = rtensor.signal_on_grid(res)
        x2 = torch.einsum('ij,baj->bai', R.T, rx2)

        ij = _find_peaks_2d(f1)
        x1p = torch.stack([x1[i, j] for i, j in ij])
        f1p = torch.stack([f1[i, j] for i, j in ij])

        ij = _find_peaks_2d(f2)
        x2p = torch.stack([x2[i, j] for i, j in ij])
        f2p = torch.stack([f2[i, j] for i, j in ij])

        # Union of the results
        mask = torch.cdist(x1p, x2p) < 2 * pi / res
        x = torch.cat([x1p[mask.sum(1) == 0], x2p])
        f = torch.cat([f1p[mask.sum(1) == 0], f2p])

        return x, f



# pylint: disable=invalid-name, arguments-differ, missing-docstring, line-too-long, no-member
import torch

from e3nn import rs


class Norm(torch.nn.Module):
    def __init__(self, Rs, normalization='component'):
        super().__init__()

        Rs = rs.simplify(Rs)
        n = sum(mul for mul, _, _ in Rs)
        self.Rs_in = Rs
        self.Rs_out = [(n, 0, +1)]
        self.normalization = normalization

    def forward(self, features):
        '''
        :param features: [..., channels]
        '''
        *size, n = features.size()
        output = []
        index = 0
        for mul, l, _ in self.Rs_in:
            sub = features.narrow(-1, index, mul * (2 * l + 1)).reshape(*size, mul, 2 * l + 1)  # [..., u, i]
            index += mul * (2 * l + 1)

            sub = sub.norm(2, dim=-1)  # [..., u]

            if self.normalization == 'component':
                sub = sub / (2 * l + 1) ** 0.5

            output.append(sub)
        assert index == n

        return torch.cat(output, dim=-1)



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, missing-docstring, arguments-differ
import numpy as np
import torch

from e3nn import rs
from e3nn.kernel_mod import FrozenKernel
from e3nn.tensor.spherical_tensor import projection


class FourierTensor:
    def __init__(self, signal, mul, lmax, p_val=0, p_arg=0):
        """
        f: s2 x r -> R^N

        Rotations
        [D(g) f](x) = f(g^{-1} x)

        Parity
        [P f](x) = p_val f(p_arg x)

        f(x) = sum F^l . Y^l(x)

        This class contains the F^l

        Rotations
        [D(g) f](x) = sum [D^l(g) F^l] . Y^l(x)         (using equiv. of Y and orthogonality of D)

        Parity
        [P f](x) = sum [p_val p_arg^l F^l] . Y^l(x)     (using parity of Y)
        """
        if signal.shape[-1] != mul * (lmax + 1)**2:
            raise ValueError(
                "Last tensor dimension and Rs do not have same dimension.")

        self.signal = signal
        self.lmax = lmax
        self.mul = mul
        self.Rs = rs.convention([(mul, l, p_val * p_arg**l)
                                 for l in range(lmax + 1)])
        self.radial_model = None

    @classmethod
    def from_geometry(cls, vectors, radial_model, lmax, sum_points=True):
        """
        :param vectors: tensor of shape [..., xyz]
        :param radial_model: function of signature R+ -> R^mul
        :param lmax: maximal order of the signal
        """
        size = vectors.shape[:-1]
        vectors = vectors.reshape(-1, 3)  # [N, 3]
        radii = vectors.norm(2, -1)
        radial_functions = radial_model(radii)
        *_size, R = radial_functions.shape
        Rs = [(R, L) for L in range(lmax + 1)]
        mul_map = rs.map_mul_to_Rs(Rs)
        radial_functions = torch.einsum('nr,dr->nd',
                                        radial_functions.repeat(1, lmax + 1),
                                        mul_map)  # [N, signal]

        Ys = projection(vectors / radii.unsqueeze(-1), lmax)  # [N, l * m]
        irrep_map = rs.map_irrep_to_Rs(Rs)
        Ys = torch.einsum('nc,dc->nd', Ys, irrep_map)  # [N, l * mul * m]

        signal = Ys * radial_functions  # [N, l * mul * m]

        if sum_points:
            signal = signal.sum(0)
        else:
            signal = signal.reshape(*size, -1)

        new_cls = cls(signal, R, lmax)
        new_cls.radial_model = radial_model
        return new_cls

    def plotly_surface(self, box_length, center=None, n=30,
                       radial_model=None, relu=True):
        """
        To use as follow
        ```
        import plotly.graph_objects as go
        surface = go.Surface(**self.plotly_surface())
        fig = go.Figure(data=[surface])
        fig.show()
        ```
        """
        r, f = self.plot(box_length, center, n, radial_model, relu)
        return dict(
            x=r[:, :, 0].numpy(),
            y=r[:, :, 1].numpy(),
            z=r[:, :, 2].numpy(),
            surfacecolor=f.numpy(),
        )

    def plot(self, box_length, center=None, n=30,
             radial_model=None, relu=True):
        muls, _ls, _ps = zip(*self.Rs)
        # We assume radial functions are repeated across L's
        assert len(set(muls)) == 1
        num_L = len(self.Rs)
        if radial_model is None:
            radial_model = self.radial_model

        def new_radial(x):
            return radial_model(x).repeat(1, num_L)  # Repeat along filter dim
        r, f = plot_on_grid(box_length, new_radial, self.Rs, n=n)
        # Multiply coefficients
        f = torch.einsum('xd,d->x', f, self.signal)
        f = f.relu() if relu else f

        if center is not None:
            r += center.unsqueeze(0)

        return r, f

    def change_lmax(self, lmax):
        new_Rs = [(self.mul, l) for l in range(lmax + 1)]
        if self.lmax == lmax:
            return self
        elif self.lmax > lmax:
            new_signal = self.signal[:rs.dim(new_Rs)]
            return FourierTensor(new_signal, self.mul, lmax)
        elif self.lmax < lmax:
            new_signal = torch.zeros(rs.dim(new_Rs))
            new_signal[:rs.dim(self.Rs)] = self.signal
            return FourierTensor(new_signal, self.mul, lmax)

    def __add__(self, other):
        if self.mul != other.mul:
            raise ValueError("Multiplicities do not match.")
        lmax = max(self.lmax, other.lmax)
        new_self = self.change_lmax(lmax)
        new_other = other.change_lmax(lmax)
        return FourierTensor(new_self.signal + new_other.signal, self.mul, self.lmax)


def plot_on_grid(box_length, radial_model, Rs, n=30):
    l_to_index = {}
    set_of_l = set([l for mul, l, p in Rs])
    start = 0
    for l in set_of_l:
        l_to_index[l] = [start, start + 2 * l + 1]
        start += 2 * l + 1

    r = np.mgrid[-1:1:n * 1j, -1:1:n * 1j, -1:1:n * 1j].reshape(3, -1)
    r = r.transpose(1, 0)
    r *= box_length / 2.
    r = torch.from_numpy(r)

    Rs_in = [(1, 0)]
    Rs_out = Rs

    def radial_lambda(_ignored):
        return radial_model

    grid = FrozenKernel(Rs_in, Rs_out, radial_lambda, r)
    f = grid()
    f = f[..., 0]
    return r, f



# pylint: disable=no-member, arguments-differ, redefined-builtin, missing-docstring, line-too-long, invalid-name, abstract-method
import torch

from e3nn import rs
from e3nn.non_linearities.rescaled_act import swish, tanh
from e3nn.non_linearities.s2 import S2Activation
from e3nn.tensor_product import LearnableTensorSquare


class S2Network(torch.nn.Module):
    def __init__(self, Rs_in, mul, lmax, Rs_out, layers=3):
        super().__init__()

        Rs = self.Rs_in = rs.simplify(Rs_in)
        self.Rs_out = rs.simplify(Rs_out)
        self.act = S2Activation(list(range(lmax + 1)),
                                swish, res=20 * (lmax + 1))

        self.layers = []

        for _ in range(layers):
            lin = LearnableTensorSquare(
                Rs, mul * self.act.Rs_in, linear=True, allow_zero_outputs=True)

            # s2 nonlinearity
            Rs = mul * self.act.Rs_out

            self.layers += [lin]

        self.layers = torch.nn.ModuleList(self.layers)

        self.tail = LearnableTensorSquare(Rs, self.Rs_out)

    def forward(self, x):
        for lin in self.layers:
            x = lin(x)

            # put multiplicity into batch
            x = x.reshape(*x.shape[:-1], -1, rs.dim(self.act.Rs_in))
            x = self.act(x)
            x = x.reshape(*x.shape[:-2], -1)  # put back into representation

        x = self.tail(x)
        return x


class S2ParityNetwork(torch.nn.Module):
    def __init__(self, Rs_in, mul, lmax, Rs_out, layers=3):
        super().__init__()

        Rs = self.Rs_in = rs.simplify(Rs_in)
        self.Rs_out = rs.simplify(Rs_out)

        def make_act(p_val, p_arg, act):
            Rs = [(1, l, p_val * p_arg**l) for l in range(lmax + 1)]
            return S2Activation(Rs, act, res=20 * (lmax + 1))

        self.act1, self.act2 = make_act(1, -1, swish), make_act(-1, -1, tanh)
        self.mul = mul

        self.layers = []

        for _ in range(layers):
            Rs_out = mul * (self.act1.Rs_in + self.act2.Rs_in)
            lin = LearnableTensorSquare(
                Rs, Rs_out, linear=True, allow_zero_outputs=True)

            # s2 nonlinearity
            Rs = mul * (self.act1.Rs_out + self.act2.Rs_out)

            self.layers += [lin]

        self.layers = torch.nn.ModuleList(self.layers)

        self.tail = LearnableTensorSquare(Rs, self.Rs_out)

    def forward(self, x):
        for lin in self.layers:
            x = lin(x)

            # put multiplicity into batch
            x = x.reshape(*x.shape[:-1], self.mul, -1)
            x1 = x.narrow(-1, 0, rs.dim(self.act1.Rs_in))
            x2 = x.narrow(-1, rs.dim(self.act1.Rs_in), rs.dim(self.act2.Rs_in))
            x1 = self.act1(x1)
            x2 = self.act2(x2)
            x = torch.cat([x1, x2], dim=-1)
            x = x.reshape(*x.shape[:-2], -1)  # put back into representation

        x = self.tail(x)
        return x



# pylint: disable=arguments-differ, redefined-builtin, missing-docstring, no-member, invalid-name, line-too-long, not-callable, abstract-method
import math
import torch
import torch_geometric as tg

from e3nn import rsh, rs
from e3nn.tensor_product import WeightedTensorProduct, GroupedWeightedTensorProduct
from e3nn.linear import Linear


class Convolution(tg.nn.MessagePassing):
    def __init__(self, kernel):
        super(Convolution, self).__init__(aggr='add', flow='target_to_source')
        self.kernel = kernel

    def forward(self, features, edge_index, edge_r, size=None, n_norm=1, groups=1):
        """
        :param features: Tensor of shape [n_target, dim(Rs_in)]
        :param edge_index: LongTensor of shape [2, num_messages]
                           edge_index[0] = sources (convolution centers)
                           edge_index[1] = targets (neighbors)
        :param edge_r: Tensor of shape [num_messages, 3]
                       edge_r = position_target - position_source
        :param size: (n_target, n_source) or None
        :param n_norm: typical number of targets per source

        :return: Tensor of shape [n_source, dim(Rs_out)]
        """
        k = self.kernel(edge_r)
        k.div_(n_norm ** 0.5)
        return self.propagate(edge_index, size=size, x=features, k=k, groups=groups)

    def message(self, x_j, k, groups):
        N = x_j.shape[0]
        cout, cin = k.shape[-2:]
        x_j = x_j.view(N, groups, cin)  # Rs_tp1
        if k.shape[0] == 0:  # https://github.com/pytorch/pytorch/issues/37628
            return torch.zeros(0, groups * cout)
        if k.dim() == 4 and k.shape[1] == groups:  # kernel has group dimension
            return torch.einsum('egij,egj->egi', k, x_j).reshape(N, groups * cout)
        return torch.einsum('eij,egj->egi', k, x_j).reshape(N, groups * cout)


class WTPConv(tg.nn.MessagePassing):
    def __init__(self, Rs_in, Rs_out, Rs_sh, RadialModel, normalization='component'):
        """
        :param Rs_in:  input representation
        :param lmax:   spherical harmonic representation
        :param Rs_out: output representation
        :param RadialModel: model constructor
        """
        super().__init__(aggr='add', flow='target_to_source')
        self.Rs_in = rs.simplify(Rs_in)
        self.Rs_out = rs.simplify(Rs_out)

        self.tp = WeightedTensorProduct(Rs_in, Rs_sh, Rs_out, normalization, own_weight=False)
        self.rm = RadialModel(self.tp.nweight)
        self.Rs_sh = Rs_sh
        self.normalization = normalization

    def forward(self, features, edge_index, edge_r, sh=None, size=None, n_norm=1):
        """
        :param features: Tensor of shape [n_target, dim(Rs_in)]
        :param edge_index: LongTensor of shape [2, num_messages]
                           edge_index[0] = sources (convolution centers)
                           edge_index[1] = targets (neighbors)
        :param edge_r: Tensor of shape [num_messages, 3]
                       edge_r = position_target - position_source
        :param sh: Tensor of shape [num_messages, dim(Rs_sh)]
        :param size: (n_target, n_source) or None
        :param n_norm: typical number of targets per source

        :return: Tensor of shape [n_source, dim(Rs_out)]
        """
        if sh is None:
            sh = rsh.spherical_harmonics_xyz(self.Rs_sh, edge_r, self.normalization)  # [num_messages, dim(Rs_sh)]
        sh = sh / n_norm**0.5

        w = self.rm(edge_r.norm(dim=1))  # [num_messages, nweight]

        return self.propagate(edge_index, size=size, x=features, sh=sh, w=w)

    def message(self, x_j, sh, w):
        """
        :param x_j: [num_messages, dim(Rs_in)]
        :param sh:  [num_messages, dim(Rs_sh)]
        :param w:   [num_messages, nweight]
        """
        return self.tp(x_j, sh, w)


class WTPConv2(tg.nn.MessagePassing):
    r"""
    WTPConv with self interaction and grouping

    This class assumes that the input and output atom positions are the same
    """
    def __init__(self, Rs_in, Rs_out, Rs_sh, RadialModel, groups=math.inf, normalization='component'):
        super().__init__(aggr='add', flow='target_to_source')
        self.Rs_in = rs.simplify(Rs_in)
        self.Rs_out = rs.simplify(Rs_out)

        self.lin1 = Linear(Rs_in, Rs_out, allow_unused_inputs=True, allow_zero_outputs=True)
        self.tp = GroupedWeightedTensorProduct(Rs_in, Rs_sh, Rs_out, groups=groups, normalization=normalization, own_weight=False)
        self.rm = RadialModel(self.tp.nweight)
        self.lin2 = Linear(Rs_out, Rs_out)
        self.Rs_sh = Rs_sh
        self.normalization = normalization

    def forward(self, features, edge_index, edge_r, sh=None, size=None, n_norm=1):
        # features = [num_atoms, dim(Rs_in)]
        if sh is None:
            sh = rsh.spherical_harmonics_xyz(self.Rs_sh, edge_r, self.normalization)  # [num_messages, dim(Rs_sh)]
        sh = sh / n_norm**0.5

        w = self.rm(edge_r.norm(dim=1))  # [num_messages, nweight]

        self_interation = self.lin1(features)
        features = self.propagate(edge_index, size=size, x=features, sh=sh, w=w)
        features = self.lin2(features)
        has_self_interaction = torch.cat([
            features.new_ones(mul * (2 * l + 1)) if any(l_in == l and p_in == p for _, l_in, p_in in self.Rs_in) else features.new_zeros(mul * (2 * l + 1))
            for mul, l, p in self.Rs_out
        ])
        return 0.5**0.5 * self_interation + (1 + (0.5**0.5 - 1) * has_self_interaction) * features

    def message(self, x_j, sh, w):
        return self.tp(x_j, sh, w)



# pylint: disable=no-member, missing-docstring, invalid-name, redefined-builtin, arguments-differ, line-too-long, unbalanced-tuple-unpacking
import torch

from e3nn import rs
from e3nn.non_linearities.norm import Norm


class NormActivation(torch.nn.Module):
    def __init__(self, Rs, activation, normalization='component'):
        super().__init__()

        self.Rs = rs.convention(Rs)
        self.activation = activation
        self.norm = Norm(self.Rs, normalization)
        self.bias = torch.nn.Parameter(torch.zeros(rs.mul_dim(self.Rs)))


    def forward(self, features):
        """
        :param features: tensor [..., channel]
        :return:         tensor [..., channel]
        """
        *size, d = features.shape
        assert d == rs.dim(self.Rs)

        norms = self.norm(features)  # [..., l*mul]

        output = []
        index_features = 0
        index_norms = 0
        for mul, l, _ in self.Rs:
            v = features.narrow(-1, index_features, mul * (2 * l + 1)).reshape(*size, mul, 2 * l + 1)  # [..., u, i]
            index_features += mul * (2 * l + 1)

            n = norms.narrow(-1, index_norms, mul).reshape(*size, mul, 1)  # [..., u, i]
            b = self.bias[index_norms: index_norms + mul].reshape(mul, 1)  # [u, i]
            index_norms += mul

            if l == 0:
                out = self.activation(v + b)
            else:
                out = self.activation(n + b) * v

            output.append(out.reshape(*size, mul * (2 * l + 1)))

        return torch.cat(output, dim=-1)



# pylint: disable=invalid-name, arguments-differ, missing-docstring, line-too-long, no-member, unbalanced-tuple-unpacking, abstract-method
import torch

from e3nn import rs
from e3nn.non_linearities.activation import Activation


class GatedBlockParity(torch.nn.Module):
    def __init__(self, Rs_scalars, act_scalars, Rs_gates, act_gates, Rs_nonscalars):
        super().__init__()

        self.Rs_in = Rs_scalars + Rs_gates + Rs_nonscalars
        self.Rs_scalars, self.Rs_gates, self.Rs_nonscalars = Rs_scalars, Rs_gates, Rs_nonscalars

        self.act_scalars = Activation(Rs_scalars, act_scalars)
        Rs_scalars = self.act_scalars.Rs_out

        self.act_gates = Activation(Rs_gates, act_gates)
        Rs_gates = self.act_gates.Rs_out

        self.mul = rs.ElementwiseTensorProduct(Rs_nonscalars, Rs_gates)
        Rs_nonscalars = self.mul.Rs_out

        self.Rs_out = Rs_scalars + Rs_nonscalars

    def __repr__(self):
        return "{name} ({Rs_scalars} + {Rs_gates} + {Rs_nonscalars} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_scalars=rs.format_Rs(self.Rs_scalars),
            Rs_gates=rs.format_Rs(self.Rs_gates),
            Rs_nonscalars=rs.format_Rs(self.Rs_nonscalars),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def forward(self, features, dim=-1, groups=None):
        scalars, gates, nonscalars = rs.cut(features, self.Rs_scalars, self.Rs_gates, self.Rs_nonscalars, dim_=dim)
        scalars = self.act_scalars(scalars)
        if gates.shape[dim] and groups is None:
            gates = self.act_gates(gates)
            nonscalars = self.mul(nonscalars, gates)
            features = torch.cat([scalars, nonscalars], dim=dim)
        # If group structure needs to be preserved
        elif gates.shape[dim] and groups is not None:
            gates = self.act_gates(gates)
            nonscalars = self.mul(nonscalars, gates)
            shape = scalars.shape
            scalars = scalars.reshape(*shape[:dim], groups, -1, *shape[dim:][1:])
            nonscalars = nonscalars.reshape(*shape[:dim], groups, -1, *shape[dim:][1:])
            features = torch.cat([scalars, nonscalars], dim=len(shape[:dim]) + 1)
            features = features.reshape(*shape[:dim], -1, *shape[dim:][1:])
        else:
            features = scalars
        return features



# pylint: disable=no-member, missing-docstring, invalid-name, arguments-differ
"""
rescaled activation functions such that the second moment is equal to one

int Dx f(x)^2 = 1

where Dx is the gaussian measure
"""
import torch


class Softplus:
    def __init__(self, beta):
        x = torch.randn(100_000, dtype=torch.float64)
        self.factor = torch.nn.functional.softplus(x, beta).pow(2).mean().rsqrt().item()
        self.beta = beta

    def __call__(self, x):
        return torch.nn.functional.softplus(x, self.beta).mul(self.factor)


softplus = Softplus(1)


class ShiftedSoftplus:
    def __init__(self, beta):
        x = torch.randn(100_000, dtype=torch.float64)
        self.shift = torch.nn.functional.softplus(torch.zeros(()), beta).item()
        y = torch.nn.functional.softplus(x, beta).sub(self.shift)
        self.factor = y.pow(2).mean().rsqrt().item()
        self.beta = beta

    def __call__(self, x):
        return torch.nn.functional.softplus(x, self.beta).sub(self.shift).mul(self.factor)


shiftedsoftplus = ShiftedSoftplus(1)


def identity(x):
    return x


def quadratic(x):
    return x.pow(2)


def sigmoid(x):
    return x.sigmoid().mul(1.84623)


def tanh(x):
    return x.tanh().mul(1.59254)


def relu(x):
    return x.relu().mul(2 ** 0.5)


def absolute(x):
    return x.abs()


@torch.jit.script
def swish_jit_fwd(x):
    return x * torch.sigmoid(x) * 1.6768


@torch.jit.script
def swish_jit_bwd(x, grad_output):
    x_sigmoid = torch.sigmoid(x)
    return grad_output * (x_sigmoid * (1 + x * (1 - x_sigmoid))) * 1.6768


class SwishJitAutoFn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return swish_jit_fwd(x)

    @staticmethod
    def backward(ctx, grad_output):  # pragma: no cover
        x = ctx.saved_tensors[0]
        return swish_jit_bwd(x, grad_output)


def swish(x):
    return SwishJitAutoFn.apply(x)



# pylint: disable=arguments-differ, redefined-builtin, missing-docstring, no-member, invalid-name, line-too-long, not-callable, abstract-method
import torch


class Convolution(torch.nn.Module):
    def __init__(self, kernel):
        super().__init__()
        self.kernel = kernel

    def forward(self, features, geometry, out_geometry=None, n_norm=1,
                custom_backward_conv=False, custom_backward_kernel=False, r_eps=0, groups=1):
        """
        :param features:     tensor [batch,  in_point, groups * channel]
        :param geometry:     tensor [batch,  in_point, xyz]
        :param out_geometry: tensor [batch, out_point, xyz]
        :param n_norm: Divide kernel by sqrt(n_norm) before passing to convolution.
        :param custom_backward_conv: call ConvolutionEinsumFn rather than using automatic differentiation
        :param custom_backward_kernel: call ConvolutionEinsumFn rather than using automatic differentiation
        :return:             tensor [batch, out_point, channel]
        """
        assert features.size()[:2] == geometry.size()[:2], "features size ({}) and geometry size ({}) should match".format(features.size(), geometry.size())
        if out_geometry is None:
            out_geometry = geometry
        rb = geometry.unsqueeze(1)  # [batch, 1, b, xyz]
        ra = out_geometry.unsqueeze(2)  # [batch, a, 1, xyz]
        k = self.kernel(rb - ra, custom_backward=custom_backward_kernel, r_eps=r_eps)  # [batch, a, b, i, j]
        k.div_(n_norm ** 0.5)

        features = features.reshape(*features.shape[:2], groups, features.shape[2] // groups)  # [batch, b, goup, j]

        if custom_backward_conv:
            features = ConvolutionEinsumFn.apply(k, features)  # [batch, point, groups, channel]
        elif k.dim() == 6 and k.shape[-3] == groups:
            features = torch.einsum("zabgij,zbgj->zagi", k, features)
        else:
            features = torch.einsum("zabij,zbgj->zagi", k, features)  # [batch, point, groups, channel]

        return features.reshape(*features.shape[:2], groups * features.shape[3])  # [batch, point, groups * channel]


class ConvolutionEinsumFn(torch.autograd.Function):
    """Forward and backward written explicitly for the Convolution Function."""
    @staticmethod
    def forward(ctx, k, features):
        """
        :param k:        tensor [batch, out_point, in_point, l_out * mul_out * m_out, l_in * mul_in * m_in]
        :param features: tensor [batch, in_point, group, l_in * mul_in * m_in]
        """
        ctx.save_for_backward(k, features)
        return torch.einsum("zabij,zbgj->zagi", k, features)  # [batch, point, channel]

    @staticmethod
    def backward(ctx, grad_output):  # pragma: no cover
        k, features = ctx.saved_tensors
        del ctx
        grad_k = torch.einsum("zagi,zbgj->zabij", grad_output, features)
        grad_features = torch.einsum("zabij,zai->zbj", k, grad_output)
        return grad_k, grad_features


class ApplyKernel(torch.nn.Module):
    def __init__(self, kernel):
        super().__init__()
        self.kernel = kernel

    def forward(self, features, geometry):
        """
        :param features: tensor [batch, point, channel]
        :param geometry: tensor [batch, point, xyz]
        :return:         tensor [batch, point, point, channel]
        """
        assert features.size()[:2] == geometry.size()[:2], "features size ({}) and geometry size ({}) should match".format(features.size(), geometry.size())
        rb = geometry.unsqueeze(1)  # [batch, 1, b, xyz]
        ra = geometry.unsqueeze(2)  # [batch, a, 1, xyz]
        k = self.kernel(rb - ra)  # [batch, a, b, i, j]
        return torch.einsum("zabij,zbj->zabi", (k, features))  # [batch, point, point, channel]



# pylint: disable=C,R,E1101,E1102
"""
Compute exact minimum bounding sphere of a 3D point cloud using Welzl's algorithm.

code copied from https://github.com/shrx/mbsc

simplified, remove colinearity checks, added naive tests in fit_sphere
"""
import numpy as np
from scipy.spatial import ConvexHull  # pylint: disable=E
from scipy.spatial.qhull import QhullError  # pylint: disable=E
from scipy import linalg


def bounding_sphere(array, eps):
    array = np.array(array)
    array = np.unique(array, axis=0)

    try:
        hull = ConvexHull(array)
        array = array[hull.vertices]
    except QhullError:
        pass

    array = np.random.permutation(array)

    if len(array) <= 4:
        return fit_sphere(array, eps)

    return B_min_sphere(list(array), [], eps)


def B_min_sphere(P, B, eps):
    if len(B) == 4 or len(P) == 0:
        return fit_sphere(B, eps)

    # Remove the last (i.e., end) point, p, from the list
    p = P[-1]
    P = P[:-1]

    # Check if p is on or inside the bounding sphere. If not, it must be part of the new boundary.
    radius, center = B_min_sphere(P, B, eps)

    if np.isnan(radius) or np.isinf(radius) or radius < eps or np.linalg.norm(p - center) > (radius + eps):
        radius, center = B_min_sphere(P, B + [p], eps)

    return radius, center


def fit_sphere(array, eps):
    """Fit a sphere to a set of 2, 3, or at most 4 points in 3D space."""

    array = np.array(array)
    N = len(array)

    # Empty set
    if N == 0:
        R = np.nan
        C = np.full(3, np.nan)
        return R, C

    # A single point
    elif N == 1:
        R = 0.
        C = array[0]
        return R, C

    # Line segment
    elif N == 2:
        R = np.linalg.norm(array[1] - array[0]) / 2
        C = np.mean(array, axis=0)
        return R, C

    elif N == 3:
        R, C = fit_sphere(array[[1, 2]], eps)
        if np.linalg.norm(array[0] - C) <= (R + eps):
            return R, C
        R, C = fit_sphere(array[[0, 2]], eps)
        if np.linalg.norm(array[1] - C) <= (R + eps):
            return R, C
        R, C = fit_sphere(array[[0, 1]], eps)
        if np.linalg.norm(array[2] - C) <= (R + eps):
            return R, C

        D12 = array[1] - array[0]
        D12 = D12 / np.linalg.norm(D12)
        D13 = array[2] - array[0]
        D13 = D13 / np.linalg.norm(D13)

        # Make plane formed by the points parallel with the xy-plane
        n = np.cross(D13, D12)
        n = n / np.linalg.norm(n)
        r = np.cross(n, np.array([0, 0, 1]))
        if np.linalg.norm(r) > 0:
            r = np.arccos(n[2]) * r / np.linalg.norm(r)  # Euler rotation vector

        Rmat = linalg.expm(np.array([
            [0., -r[2], r[1]],
            [r[2], 0., -r[0]],
            [-r[1], r[0], 0.]
        ]))

        Xr = np.transpose(np.dot(Rmat, np.transpose(array)))

        # Circle centroid
        x = Xr[:, :2]
        A = 2 * (x[1:] - np.full(2, x[0]))
        b = np.sum((np.square(x[1:]) - np.square(np.full(2, x[0]))), axis=1)
        C = np.transpose(np.linalg.solve(A, b))

        # Circle radius
        R = np.sqrt(np.sum(np.square(x[0] - C)))

        # Rotate centroid back into the original frame of reference
        C = np.append(C, [np.mean(Xr[:, 2])], axis=0)
        C = np.transpose(np.dot(np.transpose(Rmat), C))
        return R, C

    elif N == 4:
        R, C = fit_sphere(array[[1, 2, 3]], eps)
        if np.linalg.norm(array[0] - C) <= (R + eps):
            return R, C
        R, C = fit_sphere(array[[0, 2, 3]], eps)
        if np.linalg.norm(array[1] - C) <= (R + eps):
            return R, C
        R, C = fit_sphere(array[[0, 1, 3]], eps)
        if np.linalg.norm(array[2] - C) <= (R + eps):
            return R, C
        R, C = fit_sphere(array[[0, 1, 2]], eps)
        if np.linalg.norm(array[3] - C) <= (R + eps):
            return R, C

        # Centroid of the sphere
        A = 2 * (array[1:] - np.full(len(array) - 1, array[0]))
        b = np.sum((np.square(array[1:]) - np.square(np.full(len(array) - 1, array[0]))), axis=1)
        C = np.transpose(np.linalg.solve(A, b))

        # Radius of the sphere
        R = np.sqrt(np.sum(np.square(array[0] - C), axis=0))

        return R, C

    else:
        print('Input must a N-by-3 array of point coordinates, with N<=4')



# pylint: disable=invalid-name, arguments-differ, missing-docstring, line-too-long, no-member, redefined-builtin, not-callable
import torch
from torch_sparse import SparseTensor


def register_sparse_buffer(module, name, sp):
    row, col, val = sp.coo()
    module.register_buffer("{}_row".format(name), row)
    module.register_buffer("{}_col".format(name), col)
    module.register_buffer("{}_val".format(name), val)
    module.register_buffer("{}_size".format(name), torch.tensor(sp.sparse_sizes()))


def get_sparse_buffer(module, name):
    row = getattr(module, "{}_row".format(name))
    col = getattr(module, "{}_col".format(name))
    val = getattr(module, "{}_val".format(name))
    siz = getattr(module, "{}_size".format(name))
    return SparseTensor(
        row=row,
        col=col,
        value=val,
        sparse_sizes=siz.tolist(),
    )



# pylint: disable=C,R,E1101,E1102
import torch
import torch.utils.data
import os
import scipy.io
import numpy as np
from e3nn.util.bounding_sphere import bounding_sphere
import glob
from e3nn import o3


class VoxelizeBlobs:
    def __init__(self, n_qualias, size, p):
        # for QM7 size=72 and p=0.3 works fine
        self.n_qualias = n_qualias
        self.size = size
        self.p = p
        lim = self.p * 0.5 * (self.size - 1)
        self.a = torch.linspace(-lim, lim, self.size)
        self.xx = self.a.reshape(-1, 1, 1).expand(self.size, self.size, self.size)
        self.yy = self.a.reshape(1, -1, 1).expand(self.size, self.size, self.size)
        self.zz = self.a.reshape(1, 1, -1).expand(self.size, self.size, self.size)

    def __call__(self, positions, qualias):
        fields = torch.zeros((self.n_qualias, self.size, self.size, self.size))

        for position, qualia in zip(positions, qualias):
            x, y, z = position
            x, y, z = float(x), float(y), float(z)
            density = torch.exp(-((self.xx - x)**2 + (self.yy - y)**2 + (self.zz - z)**2) / (2 * self.p**2))
            density.div_(density.sum())
            fields[qualia] += density

        return fields


def center_positions(positions):
    _radius, center = bounding_sphere([pos.numpy() for pos in positions], 1e-6)
    center = torch.tensor(center).type(torch.float32)
    return [pos - center for pos in positions]


def random_rotate_translate(positions, translation=1):
    while True:
        trans = torch.rand(3) * 2 - 1
        if trans.norm() <= 1:
            break
    rot = o3.rand_rot().type(torch.float32)
    return [rot @ pos + translation * trans for pos in positions]


class QM7(torch.utils.data.Dataset):
    url = 'http://quantum-machine.org/data/qm7.mat'
    mat_file = 'qm7.mat'

    def __init__(self, root, split, transform=None):
        self.root = os.path.expanduser(root)
        self.download()
        self.transform = transform

        data = scipy.io.loadmat(os.path.join(self.root, self.mat_file))
        indices = data['P'][split]  # shape = (5, 1433)
        self.positions = data['R'][indices]  # positions, shape = (7165, 23, 3)
        self.charges = data['Z'][indices].astype(np.int32)  # charge: 5 atom types: 1, 6, 7, 8, 16, shape = (7165, 23)
        self.energies = data['T'].flatten()[indices]  # labels: atomization energies in kcal/mol, shape = (7165, )

    def __getitem__(self, index):
        positions, charges, energy = self.positions[index], self.charges[index], self.energies[index]
        positions = positions[charges > 0]
        qualias = [{1: 0, 6: 1, 7: 2, 8: 3, 16: 4}[ch] for ch in charges[charges > 0]]
        positions = [torch.tensor(pos) for pos in positions]
        if self.transform is not None:
            return self.transform(positions, qualias, energy)
        return positions, qualias, energy

    def __len__(self):
        return len(self.energies)

    def download(self):
        if not os.path.isdir(self.root):
            os.makedirs(self.root)

        file_path = os.path.join(self.root, self.mat_file)
        if not os.path.isfile(file_path):
            print('Download...')
            from six.moves import urllib
            data = urllib.request.urlopen(self.url)
            with open(file_path, 'wb') as f:
                f.write(data.read())


class QM9(torch.utils.data.Dataset):
    url = 'https://ndownloader.figshare.com/files/3195389'
    properties_names = [
        'A GHz Rotational constant',
        'B GHz Rotational constant',
        'C GHz Rotational constant',
        'μ D Dipole moment',
        'α a^3_0 Isotropic polarizability',
        'ϵ_HOMO Ha Energy of HOMO',
        'ϵ_LUMO Ha Energy of LUMO',
        'ϵ_gap Ha Gap (ϵLUMO−ϵHOMO)',
        'R^2 a^2_0 Electronic spatial extent',
        'zpve Ha Zero point vibrational energy',
        'U_0 Ha Internal energy at 0K',
        'U Ha Internal energy at 298.15K',
        'H Ha Enthalpy at 298.15K',
        'G Ha Free energy at 298.15K',
        'C_v cal/(molK) Heat capacity at 298.15K',
    ]

    def __init__(self, root, transform=None):
        self.root = os.path.expanduser(root)
        self.files = None
        self.download()
        self.transform = transform

    def __getitem__(self, index):
        path = self.files[index]
        with open(path, "rt") as f:
            lines = f.readlines()
        n = int(lines[0])
        properties = [float(x) for x in lines[1].split()[2:]]
        # https://www.nature.com/articles/sdata201422/tables/4
        qualias = []
        positions = []
        for i in range(2, 2 + n):
            atom, x, y, z, _ = lines[i].split()
            qualias.append("CHONF".index(atom))
            positions.append([float(a.replace('*^', 'e')) for a in [x, y, z]])
        positions = [torch.tensor(pos) for pos in positions]
        if self.transform is not None:
            return self.transform(positions, qualias, properties)
        return positions, qualias, properties

    def __len__(self):
        return len(self.files)

    def download(self):
        if not os.path.isdir(self.root):
            os.makedirs(self.root)

        bz2_path = os.path.join(self.root, "data.xyz.tar.bz2")
        if not os.path.isfile(bz2_path):
            print("Download...")
            from six.moves import urllib
            data = urllib.request.urlopen(self.url)
            with open(bz2_path, 'wb') as f:
                f.write(data.read())

        tar_path = os.path.join(self.root, "data.xyz.tar")
        if not os.path.isfile(tar_path):
            print("Decompress...")
            import bz2
            file = bz2.BZ2File(bz2_path)
            with open(tar_path, 'wb') as f:
                f.write(file.read())
            file.close()

        xyz_path = os.path.join(self.root, "data.xyz")
        if not os.path.isdir(xyz_path):
            print("Extract...")
            import tarfile
            tar = tarfile.open(tar_path)
            tar.extractall(xyz_path)
            tar.close()

        self.files = sorted(glob.glob(os.path.join(xyz_path, "*.xyz")))
        assert len(self.files) == 133885



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, wildcard-import, unused-wildcard-import, missing-docstring
from typing import Tuple, Set
import random
import math

TY_PERM = Tuple[int]


def is_perm(p: TY_PERM):
    return sorted(set(p)) == list(range(len(p)))


def identity(n: int) -> TY_PERM:
    return tuple(i for i in range(n))


def compose(p1: TY_PERM, p2: TY_PERM) -> TY_PERM:
    """
    compute p1 . p2
    """
    assert is_perm(p1) and is_perm(p2)
    assert len(p1) == len(p2)
    # p: i |-> p[i]

    # [p1.p2](i) = p1(p2(i)) = p1[p2[i]]
    return tuple(p1[p2[i]] for i in range(len(p1)))


def inverse(p: TY_PERM) -> TY_PERM:
    """
    compute the inverse permutation
    """
    return tuple(p.index(i) for i in range(len(p)))


def rand(n: int) -> TY_PERM:
    i = random.randint(0, math.factorial(n) - 1)
    return from_int(i, n)


def from_int(i: int, n: int) -> TY_PERM:
    pool = list(range(n))
    p = []
    for _ in range(n):
        j = i % n
        i = i // n
        p.append(pool.pop(j))
        n -= 1
    return tuple(p)


def to_int(p: TY_PERM) -> int:
    n = len(p)
    pool = list(range(n))
    i = 0
    m = 1
    for j in p:
        k = pool.index(j)
        i += k * m
        m *= len(pool)
        pool.pop(k)
    return i


def group(n: int) -> Set[TY_PERM]:
    return {from_int(i, n) for i in range(math.factorial(n))}


def germinate(subset: Set[TY_PERM]) -> Set[TY_PERM]:
    while True:
        n = len(subset)
        subset = subset.union([inverse(p) for p in subset])
        subset = subset.union([
            compose(p1, p2)
            for p1 in subset
            for p2 in subset
        ])
        if len(subset) == n:
            return subset


def is_group(g: Set[TY_PERM]) -> bool:
    if len(g) == 0:
        return False

    n = len(next(iter(g)))

    for p in g:
        assert len(p) == n, p

    if not identity(n) in g:
        return False

    for p in g:
        if not inverse(p) in g:
            return False

    for p1 in g:
        for p2 in g:
            if not compose(p1, p2) in g:
                return False

    return True



# pylint: disable=invalid-name, arguments-differ, missing-docstring, no-member
import copy

import torch

from e3nn import rs


class Activation(torch.nn.Module):
    def __init__(self, Rs, acts):
        '''
        Can be used only with scalar fields

        :param acts: list of tuple (multiplicity, activation)
        '''
        super().__init__()

        Rs = rs.simplify(Rs)
        acts = copy.deepcopy(acts)

        n1 = sum(mul for mul, _, _ in Rs)
        n2 = sum(mul for mul, _ in acts if mul > 0)

        for i, (mul, act) in enumerate(acts):
            if mul == -1:
                acts[i] = (n1 - n2, act)
                assert n1 - n2 >= 0

        assert n1 == sum(mul for mul, _ in acts)

        i = 0
        while i < len(Rs):
            mul_r, l, p_r = Rs[i]
            mul_a, act = acts[i]

            if mul_r < mul_a:
                acts[i] = (mul_r, act)
                acts.insert(i + 1, (mul_a - mul_r, act))

            if mul_a < mul_r:
                Rs[i] = (mul_a, l, p_r)
                Rs.insert(i + 1, (mul_r - mul_a, l, p_r))
            i += 1

        x = torch.linspace(0, 10, 256)

        Rs_out = []
        for (mul, l, p_in), (mul_a, act) in zip(Rs, acts):
            assert mul == mul_a

            a1, a2 = act(x), act(-x)
            if (a1 - a2).abs().max() < a1.abs().max() * 1e-10:
                p_act = 1
            elif (a1 + a2).abs().max() < a1.abs().max() * 1e-10:
                p_act = -1
            else:
                p_act = 0

            p = p_act if p_in == -1 else p_in
            Rs_out.append((mul, 0, p))

            if p_in != 0 and p == 0:
                raise ValueError("warning! the parity is violated")

        self.Rs_out = Rs_out
        self.acts = acts

    def forward(self, features, dim=-1):
        '''
        :param features: [..., channels, ...]
        '''
        output = []
        index = 0
        for mul, act in self.acts:
            output.append(act(features.narrow(dim, index, mul)))
            index += mul

        if output:
            return torch.cat(output, dim=dim)
        else:
            size = list(features.size())
            size[dim] = 0
            return features.new_zeros(*size)
